import math
from typing import Sequence

import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn

from resselt.utilities import DySample


class Interpolate(nn.Module):
    def __init__(self, scale_factor: int = 4, mode: str = 'nearest'):
        super(Interpolate, self).__init__()
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self, x):
        return F.interpolate(x, scale_factor=self.scale_factor, mode=self.mode)


class InterpolateUpsampler(nn.Sequential):
    def __init__(self, dim: int = 64, out_ch: int = 3, scale: int = 4):
        m = []
        if (scale & (scale - 1)) == 0:
            for _ in range(int(math.log2(scale))):
                m.append(nn.Conv2d(dim, dim, 3, 1, 1))
                m.append(Interpolate(2))
                m.append(nn.LeakyReLU(negative_slope=0.2, inplace=True))
            m.append(nn.Conv2d(dim, dim, 3, 1, 1))
            m.append(nn.LeakyReLU(negative_slope=0.2, inplace=True))
        elif scale == 3:
            m.append(nn.Conv2d(dim, dim, 3, 1, 1))
            m.append(Interpolate(scale))
            m.append(nn.LeakyReLU(negative_slope=0.2, inplace=True))
            m.append(nn.Conv2d(dim, dim, 3, 1, 1))
            m.append(nn.LeakyReLU(negative_slope=0.2, inplace=True))

        m.append(nn.Conv2d(dim, out_ch, 3, 1, 1))
        super().__init__(*m)


class ConvBlock(nn.Module):
    r"""https://github.com/joshyZhou/AST/blob/main/model.py#L22"""

    def __init__(self, in_channel: int = 3, out_channel: int = 48, strides: int = 1):
        super().__init__()
        self.strides = strides
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.block = nn.Sequential(
            nn.Conv2d(in_channel, out_channel, kernel_size=3, stride=strides, padding=1),
            nn.Mish(),
            nn.Conv2d(out_channel, out_channel, kernel_size=3, stride=strides, padding=1),
            nn.Mish(),
        )
        self.conv11 = nn.Conv2d(in_channel, out_channel, kernel_size=1, stride=strides, padding=0)

    def forward(self, x):
        out1 = self.block(x)
        out2 = self.conv11(x)
        return out1 + out2


class OmniShift(nn.Module):
    def __init__(self, dim: int = 48):
        super(OmniShift, self).__init__()
        # Define the layers for training
        self.conv1x1 = nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=1, groups=dim, bias=False)
        self.conv3x3 = nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=3, padding=1, groups=dim, bias=False)
        self.conv5x5 = nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=5, padding=2, groups=dim, bias=False)
        self.alpha = nn.Parameter(torch.randn(4), requires_grad=True)

        # Define the layers for testing
        self.conv5x5_reparam = nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=5, padding=2, groups=dim, bias=False)
        self.repram_flag = True

    def forward_train(self, x):
        out1x1 = self.conv1x1(x)
        out3x3 = self.conv3x3(x)
        out5x5 = self.conv5x5(x)
        # import pdb
        # pdb.set_trace()

        out = self.alpha[0] * x + self.alpha[1] * out1x1 + self.alpha[2] * out3x3 + self.alpha[3] * out5x5
        return out

    def reparam_5x5(self):
        # Combine the parameters of conv1x1, conv3x3, and conv5x5 to form a single 5x5 depth-wise convolution

        padded_weight_1x1 = F.pad(self.conv1x1.weight, (2, 2, 2, 2))
        padded_weight_3x3 = F.pad(self.conv3x3.weight, (1, 1, 1, 1))

        identity_weight = F.pad(torch.ones_like(self.conv1x1.weight), (2, 2, 2, 2))

        combined_weight = (
            self.alpha[0] * identity_weight
            + self.alpha[1] * padded_weight_1x1
            + self.alpha[2] * padded_weight_3x3
            + self.alpha[3] * self.conv5x5.weight
        )

        device = self.conv5x5_reparam.weight.device

        combined_weight = combined_weight.to(device)

        self.conv5x5_reparam.weight = nn.Parameter(combined_weight)

    def forward(self, x):
        # if self.training:
        #     self.repram_flag = True
        #     out = self.forward_train(x)
        # elif self.training and self.repram_flag:
        #     self.reparam_5x5()
        #     self.repram_flag = False
        #     out = self.conv5x5_reparam(x)
        # elif self.training and not self.repram_flag:
        #     out = self.conv5x5_reparam(x)
        if self.repram_flag:
            self.reparam_5x5()
            self.repram_flag = False
            out = self.conv5x5_reparam(x)
        else:
            out = self.conv5x5_reparam(x)
        return out


class MatMul(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, a, b):
        out = a @ b
        return out


class LMLTVIT(nn.Module):
    def __init__(self, dim: int = 48, window_size: int = 8, attn_drop: float = 0.0, proj_drop: float = 0.0):
        super().__init__()

        self.dim = dim
        self.window_size = window_size
        self.scale = dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.omni_shift = OmniShift(dim)
        self.get_v = nn.Conv2d(dim, dim, kernel_size=3, stride=1, padding=1, groups=dim)

    def window_partition(self, x, window_size):
        B, H, W, C = x.shape
        x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
        return x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)

    def window_reverse(self, windows, window_size, h, w):
        """
        Args:
            windows: (num_windows*b, window_size, window_size, c)
            window_size (int): Window size
            h (int): Height of image
            w (int): Width of image

        Returns:
            x: (b, h, w, c)
        """
        b = int(windows.shape[0] / (h * w / window_size / window_size))
        x = windows.view(b, h // window_size, w // window_size, window_size, window_size, -1)
        return x.permute(0, 1, 3, 2, 4, 5).contiguous().view(b, h, w, -1)

    def get_lepe(self, x, func):
        B, N, C = x.shape
        H = W = 8
        x = x.transpose(-2, -1).contiguous().view(B, C, H, W)

        H_sp, W_sp = self.window_size, self.window_size
        x = x.view(B, C, H // H_sp, H_sp, W // W_sp, W_sp)
        x = x.permute(0, 2, 4, 1, 3, 5).contiguous().reshape(-1, C, H_sp, W_sp)  # B', C, H', W'

        lepe = func(x)  # B', C, H', W'
        lepe = lepe.reshape(-1, C, H_sp * W_sp).permute(0, 2, 1).contiguous()

        x = x.reshape(-1, C, H_sp * W_sp).permute(0, 2, 1).contiguous()
        return x, lepe

    def forward(self, x, resolution):
        _B, N, C = x.shape
        H, W = resolution
        x = rearrange(x, 'b (h w) c -> b c h w', h=H, w=W)
        x = self.omni_shift(x)
        x = rearrange(x, 'b c h w -> b h w c')
        ################################
        # 1. window partition
        ################################
        x_window = self.window_partition(x, self.window_size).permute(0, 3, 1, 2)
        x_window = x_window.permute(0, 2, 3, 1).view(-1, self.window_size * self.window_size, C)

        ################################
        # 2. make qkv
        ################################
        qkv = self.qkv(x_window)
        # qkv = qkv.permute(0,2,3,1)
        # qkv = qkv.reshape(-1, self.window_size * self.window_size, 3*C)
        q, k, v = torch.chunk(qkv, 3, dim=-1)

        ################################
        # 3. attn and PE
        ################################
        v, lepe = self.get_lepe(v, self.get_v)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v) + lepe
        # x = x.reshape(-1, self.window_size, self.window_size, C)
        # x = x.permute(0,3,1,2)

        ################################
        # 4. proj and drop
        ################################
        x = self.proj(x)
        x = self.proj_drop(x)

        x = x.reshape(-1, self.window_size, self.window_size, C)
        x = self.window_reverse(x, self.window_size, H, W)
        x = rearrange(x, 'b h w c-> b (h w) c')
        return x


class ChannelMix(nn.Module):
    def __init__(self, n_embd: int = 48, hidden_rate: int = 4, key_norm: bool = False):
        super().__init__()
        self.n_embd = n_embd

        hidden_sz = int(hidden_rate * n_embd)
        self.key = nn.Linear(n_embd, hidden_sz, bias=False)

        self.omni_shift = OmniShift(dim=n_embd)

        if key_norm:
            self.key_norm = nn.RMSNorm(hidden_sz)
        else:
            self.key_norm = None
        self.receptance = nn.Linear(n_embd, n_embd, bias=False)
        self.value = nn.Linear(hidden_sz, n_embd, bias=False)

    def forward(self, x, resolution):
        h, w = resolution

        x = rearrange(x, 'b (h w) c -> b c h w', h=h, w=w)
        x = self.omni_shift(x)
        x = rearrange(x, 'b c h w -> b (h w) c')

        k = self.key(x)
        k = torch.square(torch.relu(k))
        if self.key_norm is not None:
            k = self.key_norm(k)
        kv = self.value(k)
        x = torch.sigmoid(self.receptance(x)) * kv

        return x


class TransformerBlock(nn.Module):
    def __init__(
        self, dim: int, window_size: int = 8, hidden_rate: int = 4, channel_norm: bool = False, attn_drop: float = 0.0, proj_drop: float = 0.0
    ):
        super().__init__()
        self.rn1 = nn.RMSNorm(dim)
        self.rn2 = nn.RMSNorm(dim)
        self.att = LMLTVIT(dim, window_size, attn_drop, proj_drop)
        self.ffn = ChannelMix(dim, hidden_rate, channel_norm)
        self.gamma1 = nn.Parameter(torch.ones(dim), requires_grad=True)
        self.gamma2 = nn.Parameter(torch.ones(dim), requires_grad=True)

    def forward(self, x, resolution):
        x = x + self.gamma1 * self.att(self.rn1(x), resolution)
        x = x + self.gamma2 * self.ffn(self.rn2(x), resolution)
        return x


class LBlock(nn.Module):
    def __init__(
        self,
        dim: int = 48,
        n_block: int = 1,
        window_size: int = 8,
        hidden_rate: int = 4,
        channel_norm: bool = False,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ):
        super(LBlock, self).__init__()
        self.t_blocks = nn.ModuleList([TransformerBlock(dim, window_size, hidden_rate, channel_norm, attn_drop, proj_drop) for _ in range(n_block)])
        self.conv = ConvBlock(dim * 2, dim)

    def forward(self, x, resolution):
        H, W = resolution
        shortcut = x
        for t_block in self.t_blocks:
            x = t_block(x, resolution)
        x = torch.cat([shortcut, x], dim=-1)
        x = rearrange(x, 'b (h w) c -> b c h w', h=H, w=W)
        x = self.conv(x)
        x = rearrange(x, 'b c h w -> b (h w) c')
        return x


class MBlock(nn.Module):
    def __init__(
        self,
        dim: int = 48,
        n_block: int = 1,
        window_size: int = 8,
        hidden_rate: int = 4,
        channel_norm: bool = False,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ):
        super(MBlock, self).__init__()
        self.t_blocks = nn.ModuleList([TransformerBlock(dim, window_size, hidden_rate, channel_norm, attn_drop, proj_drop) for _ in range(n_block)])
        self.conv = ConvBlock(dim * 2, dim)

    def forward(self, x):
        B, C, H, W = x.shape
        resolution = (H, W)
        x = rearrange(x, 'b c h w -> b (h w) c')
        shortcut = x
        for t_block in self.t_blocks:
            x = t_block(x, resolution)
        x = torch.cat([shortcut, x], dim=-1)
        x = rearrange(x, 'b (h w) c -> b c h w', h=H, w=W)
        x = self.conv(x)
        return x


class LinearPipeline(nn.Module):
    def __init__(
        self,
        dim: int = 48,
        num_blocks: Sequence[int] = (4, 6, 6, 8),
        window_size: int = 8,
        hidden_rate: int = 4,
        channel_norm: bool = False,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ):
        super(LinearPipeline, self).__init__()
        self.att = nn.ModuleList([LBlock(dim, num_block, window_size, hidden_rate, channel_norm, attn_drop, proj_drop) for num_block in num_blocks])

    def forward(self, x):
        _, _, H, W = x.size()
        resolution = (H, W)
        x = rearrange(x, 'b c h w -> b (h w) c')
        for attn in self.att:
            x = attn(x, resolution)
        x = rearrange(x, 'b (h w) c -> b c h w', h=H, w=W)
        return x


class Downsample(nn.Module):
    def __init__(self, n_feat: int = 48):
        super(Downsample, self).__init__()
        self.body = nn.Sequential(nn.Conv2d(n_feat, n_feat // 2, kernel_size=3, stride=1, padding=1, bias=False), nn.PixelUnshuffle(2))

    def forward(self, x):
        return self.body(x)


class Upsample(nn.Module):
    def __init__(self, n_feat: int = 48):
        super(Upsample, self).__init__()

        self.body = nn.Sequential(nn.Conv2d(n_feat, n_feat, kernel_size=3, stride=1, padding=1, bias=False), nn.PixelShuffle(2))

    def forward(self, x):
        return self.body(x)


class MetaPipeline(nn.Module):
    def __init__(
        self,
        dim: int = 48,
        num_blocks: Sequence[int] = (4, 6, 6, 8),
        window_size: int = 8,
        hidden_rate: int = 4,
        channel_norm: bool = False,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ):
        super(MetaPipeline, self).__init__()
        self.enc0 = nn.Sequential(*[MBlock(dim, num_blocks[0], window_size, hidden_rate, channel_norm, attn_drop, proj_drop)])
        self.enc1 = nn.Sequential(*[MBlock(dim * 2, num_blocks[1], window_size, hidden_rate, channel_norm, attn_drop, proj_drop)])
        self.enc2 = nn.Sequential(*[MBlock(dim * 4, num_blocks[2], window_size, hidden_rate, channel_norm, attn_drop, proj_drop)])
        self.enc3 = nn.Sequential(*[MBlock(dim * 8, num_blocks[3], window_size, hidden_rate, channel_norm, attn_drop, proj_drop)])

        self.down1 = Downsample(dim)
        self.down2 = Downsample(dim * 2)
        self.down3 = Downsample(dim * 4)

        self.up1 = Upsample(dim * 16)
        self.up2 = Upsample(dim * 8)
        self.up3 = Upsample(dim * 4)

        self.dec0 = nn.Sequential(*[MBlock(dim * 4, num_blocks[2], window_size, hidden_rate, channel_norm, attn_drop, proj_drop)])
        self.dec1 = nn.Sequential(*[MBlock(dim * 2, num_blocks[1], window_size, hidden_rate, channel_norm, attn_drop, proj_drop)])
        self.dec2 = nn.Sequential(*[MBlock(dim, num_blocks[0], window_size, hidden_rate, channel_norm, attn_drop, proj_drop)])

    def forward(self, x):
        enc0 = self.enc0(x)
        enc0 = self.down1(enc0)

        enc1 = self.enc1(enc0)
        enc1 = self.down2(enc1)

        enc2 = self.enc2(enc1)
        enc2 = self.down3(enc2)

        enc3 = self.enc3(enc2)
        enc3 = torch.cat([enc3, enc2], dim=1)

        x = self.up1(enc3)
        x = self.dec0(x)
        x = torch.cat([x, enc1], dim=1)

        x = self.up2(x)
        x = self.dec1(x)
        x = torch.cat([x, enc0], dim=1)

        x = self.up3(x)
        x = self.dec2(x)
        return x


class FlexNet(nn.Module):
    def __init__(
        self,
        inp_channels: int = 3,
        out_channels: int = 3,
        scale: int = 4,
        dim: int = 64,
        num_blocks: Sequence[int] = (6, 6, 6, 6, 6, 6),  # meta = (8,8,8,8), # linear = (6, 6, 6, 6, 6, 6),
        window_size: int = 8,
        hidden_rate: int = 4,
        channel_norm: bool = False,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        pipeline_type: str = 'linear',
        upsampler: str = 'ps',
    ):
        super(FlexNet, self).__init__()
        self.register_buffer('window_size', torch.tensor(window_size, dtype=torch.uint8))
        self.pipeline_type = pipeline_type
        self.scale = scale
        self.short_cut = ConvBlock(inp_channels, dim)
        self.in_to_feat = nn.Conv2d(inp_channels, dim, 3, 1, 1)
        self.pipeline = (
            LinearPipeline(dim, num_blocks, window_size, hidden_rate, channel_norm, attn_drop, proj_drop)
            if pipeline_type == 'linear'
            else MetaPipeline(dim, num_blocks, window_size, hidden_rate, channel_norm, attn_drop, proj_drop)
        )
        if upsampler == 'n+c':
            self.register_buffer('scale_factor', torch.tensor(scale, dtype=torch.uint8))
            self.to_img = nn.Sequential(nn.Conv2d(dim * 2, dim, 3, 1, 1), InterpolateUpsampler(dim, out_channels, scale))
        elif upsampler == 'dys':
            self.to_img = DySample(dim * 2, out_channels, scale)
        else:
            self.to_img = nn.Sequential(nn.Conv2d(dim * 2, out_channels * (scale**2), 3, 1, 1), nn.PixelShuffle(scale))

    def check_img_size(self, x, resolution):
        h, w = resolution
        scaled_size = self.window_size.to(int)
        if self.pipeline_type == 'meta':
            scaled_size *= 8
        mod_pad_h = (scaled_size - h % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - w % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, x):
        _, _, h, w = x.size()
        x = self.check_img_size(x, (h, w))
        short_cut = self.short_cut(x)
        x = self.in_to_feat(x)
        x = self.pipeline(x)
        x = torch.cat([x, short_cut], dim=1)
        x = self.to_img(x)
        return x[:, :, : h * self.scale, : w * self.scale]



import math
from typing import Mapping

from .arch import OmniSR
import warnings

from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import pixelshuffle_scale, get_seq_len

warnings.filterwarnings('ignore')


class OmniArch(Architecture[OmniSR]):
    def __init__(self):
        super().__init__(
            uid='OmniSR',
            detect=KeyCondition.has_all(
                'residual_layer.0.residual_layer.0.layer.0.fn.0.weight',
                'input.weight',
                'up.0.weight',
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        # Remove junk from the state dict
        state_dict_keys = set(state_dict.keys())
        for key in state_dict_keys:
            if key.endswith(('total_ops', 'total_params')):
                del state_dict[key]
        window_size = 8

        num_feat = state_dict['input.weight'].shape[0]
        num_in_ch = state_dict['input.weight'].shape[1]
        num_out_ch = num_in_ch
        bias = 'input.bias' in state_dict

        pixelshuffle_shape = state_dict['up.0.weight'].shape[0]
        up_scale = pixelshuffle_scale(pixelshuffle_shape, num_in_ch)

        res_num = get_seq_len(state_dict, 'residual_layer')
        block_num = get_seq_len(state_dict, 'residual_layer.0.residual_layer') - 1

        rel_pos_bias_key = 'residual_layer.0.residual_layer.0.layer.2.fn.rel_pos_bias.weight'
        if rel_pos_bias_key in state_dict:
            pe = True
            rel_pos_bias_weight = state_dict[rel_pos_bias_key].shape[0]
            window_size = int((math.sqrt(rel_pos_bias_weight) + 1) / 2)
        else:
            pe = False

        model = OmniSR(
            num_in_ch=num_in_ch,
            num_out_ch=num_out_ch,
            num_feat=num_feat,
            block_num=block_num,
            pe=pe,
            window_size=window_size,
            res_num=res_num,
            up_scale=up_scale,
            bias=bias,
        )

        return self._enhance_model(model=model, in_channels=num_in_ch, out_channels=num_out_ch, upscale=up_scale, name='OmniSR')



import torch
from torch import nn


def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    From: https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/layers/drop.py
    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    # work with diff dim tensors, not just 2D ConvNets
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).

    From: https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/layers/drop.py
    """

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)



from typing import Mapping

from .arch import GateR
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len


class GateRArch(Architecture[GateR]):
    def __init__(self):
        super().__init__(
            uid='GateR',
            detect=KeyCondition.has_all(
                'dec0.0.bias',
                'dec0.0.weight',
                'dec0.1.gated.0.conv.conv.bias',
                'dec0.1.gated.0.conv.conv.weight',
                'dec0.1.gated.0.fc1.bias',
                'dec0.1.gated.0.fc1.weight',
                'dec0.1.gated.0.fc2.bias',
                'dec0.1.gated.0.fc2.weight',
                'dec0.1.gated.0.norm.weight',
                'dec0.2.body.0.bias',
                'dec0.2.body.0.weight',
                'dec1.0.bias',
                'dec1.0.weight',
                'dec1.1.gated.0.conv.conv.bias',
                'dec1.1.gated.0.conv.conv.weight',
                'dec1.1.gated.0.fc1.bias',
                'dec1.1.gated.0.fc1.weight',
                'dec1.1.gated.0.fc2.bias',
                'dec1.1.gated.0.fc2.weight',
                'dec1.1.gated.0.norm.weight',
                'dec1.2.body.0.bias',
                'dec1.2.body.0.weight',
                'dec2.0.gated.0.conv.conv.bias',
                'dec2.0.gated.0.conv.conv.weight',
                'dec2.0.gated.0.fc1.bias',
                'dec2.0.gated.0.fc1.weight',
                'dec2.0.gated.0.fc2.bias',
                'dec2.0.gated.0.fc2.weight',
                'dec2.0.gated.0.norm.weight',
                'dim_to_ch.0.bias',
                'dim_to_ch.0.weight',
                'dim_to_ch.1.bias',
                'dim_to_ch.1.weight',
                'enc0.gated.0.conv.conv.bias',
                'enc0.gated.0.conv.conv.weight',
                'enc0.gated.0.fc1.bias',
                'enc0.gated.0.fc1.weight',
                'enc0.gated.0.fc2.bias',
                'enc0.gated.0.fc2.weight',
                'enc0.gated.0.norm.weight',
                'enc1.0.body.0.bias',
                'enc1.0.body.0.weight',
                'enc1.1.gated.0.conv.conv.bias',
                'enc1.1.gated.0.conv.conv.weight',
                'enc1.1.gated.0.fc1.bias',
                'enc1.1.gated.0.fc1.weight',
                'enc1.1.gated.0.fc2.bias',
                'enc1.1.gated.0.fc2.weight',
                'enc1.1.gated.0.norm.weight',
                'enc2.0.body.0.bias',
                'enc2.0.body.0.weight',
                'enc2.1.gated.0.conv.conv.bias',
                'enc2.1.gated.0.conv.conv.weight',
                'enc2.1.gated.0.fc1.bias',
                'enc2.1.gated.0.fc1.weight',
                'enc2.1.gated.0.fc2.bias',
                'enc2.1.gated.0.fc2.weight',
                'enc2.1.gated.0.norm.weight',
                'in_to_dim.bias',
                'in_to_dim.weight',
                'latent.0.body.0.bias',
                'latent.0.body.0.weight',
                'latent.1.gated.0.fc1.bias',
                'latent.1.gated.0.fc1.weight',
                'latent.1.gated.0.fc2.bias',
                'latent.1.gated.0.fc2.weight',
                'latent.1.gated.0.norm.weight',
                'latent.2.body.0.bias',
                'latent.2.body.0.weight',
            ),
        )

    def load(self, state: Mapping[str, object]):
        block_list = ['enc0', 'enc1.1', 'enc2.1', 'latent.1', 'dec0.1', 'dec1.1', 'dec2.0']
        dim, in_ch = state['in_to_dim.weight'].shape[:2]
        num_blocks = [get_seq_len(state, block + '.gated') for block in block_list]
        latent_att = 'latent.1.gated.0.conv.conv.weight' not in state
        model = GateR(in_ch=in_ch, dim=dim, num_blocks=num_blocks, latent_att=latent_att)

        return self._enhance_model(model=model, in_channels=in_ch, out_channels=int(in_ch), upscale=1, name='GateR')



import math
from typing import Mapping

from .arch import RHA
from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import get_seq_len


class RHAArch(Architecture[RHA]):
    def __init__(self):
        super().__init__(
            uid='RHA',
            detect=KeyCondition.has_all(
                'body.0.down_sample',
                'body.0.body.0.norm.weight',
                'body.0.body.0.norm.bias',
                'body.0.body.0.fc1.weight',
                'body.0.body.0.fc1.bias',
                'body.0.body.0.conv.att.2.scale',
                'body.0.body.0.conv.att.2.positional_encoding',
                'body.0.body.0.conv.att.2.qkv.weight',
                'body.0.body.0.conv.att.2.qkv.bias',
                'body.0.body.0.conv.att.2.proj.weight',
                'body.0.body.0.conv.att.2.proj.bias',
                'body.0.body.0.conv.att.2.dwc.weight',
                'body.0.body.0.conv.att.2.dwc.bias',
                'body.0.body.0.conv.conv.alpha1',
                'body.0.body.0.conv.conv.alpha2',
                'body.0.body.0.conv.conv.alpha3',
                'body.0.body.0.conv.conv.alpha4',
                'body.0.body.0.conv.conv.conv1x1.weight',
                'body.0.body.0.conv.conv.conv1x1.bias',
                'body.0.body.0.conv.conv.conv3x3.weight',
                'body.0.body.0.conv.conv.conv3x3.bias',
                'body.0.body.0.conv.conv.conv5x5.weight',
                'body.0.body.0.conv.conv.conv5x5.bias',
                'body.0.body.0.conv.conv.conv5x5_reparam.weight',
                'body.0.body.0.conv.conv.conv5x5_reparam.bias',
                'body.0.body.0.conv.aggr.0.weight',
                'body.0.body.0.conv.aggr.0.bias',
                'body.0.body.0.fc2.weight',
                'body.0.body.0.fc2.bias',
                'to_img.MetaUpsample',
            ),
        )

    def load(self, state: Mapping[str, object]):
        upsample = ['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']
        unshuffle = 1
        unshuffle_mod = False
        if 'unshuffle' in state:
            unshuffle = state['unshuffle'].item()
            unshuffle_mod = True
            dim, in_ch, _, _ = state['to_feat.1.weight'].shape
            in_ch //= unshuffle**2
        else:
            dim, in_ch, _, _ = state['to_feat.weight'].shape
        in_ch //= unshuffle**2
        group_blocks = get_seq_len(state, 'body')
        res_blocks = get_seq_len(state, 'body.0.body') - 2
        down_list = [state[f'body.{index}.down_sample'].item() for index in range(group_blocks)]
        expansion_ratio = state['body.0.body.0.fc1.weight'].shape[0] / 2 / dim
        _, index, scale, _, out_ch, upsample_dim, _ = [value.item() for value in state['to_img.MetaUpsample']]
        upsampler = upsample[int(index)]
        scale //= unshuffle
        window_size = math.isqrt(state['body.0.body.0.conv.att.2.positional_encoding'].shape[1])
        model = RHA(
            dim=dim,
            scale=scale,
            in_ch=in_ch,
            out_ch=out_ch,
            mid_dim=upsample_dim,
            down_list=down_list,
            expansion_ratio=expansion_ratio,
            group_blocks=group_blocks,
            res_blocks=res_blocks,
            upsample=upsampler,
            unshuffle_mod=unshuffle_mod,
            window_size=window_size,
        )

        return self._enhance_model(model=model, in_channels=in_ch, out_channels=out_ch, upscale=scale, name='RHA')



import math
from typing import Mapping

from .arch import FlexNet
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len


class FlexNetArch(Architecture[FlexNet]):
    def __init__(self):
        super().__init__(
            uid='FlexNet',
            detect=KeyCondition.has_all(
                'short_cut.block.0.weight',
                'short_cut.block.0.bias',
                'short_cut.block.2.weight',
                'short_cut.block.2.bias',
                'short_cut.conv11.weight',
                'short_cut.conv11.bias',
                'in_to_feat.weight',
                'in_to_feat.bias',
                KeyCondition.has_any(
                    'pipeline.enc0.0.t_blocks.0.gamma1',
                    'pipeline.att.0.t_blocks.0.gamma1',
                ),
            ),
        )

    def load(self, state: Mapping[str, object]):
        window_size = int(state['window_size'])
        dim, inp_channels = state['in_to_feat.weight'].shape[:2]
        out_channels = inp_channels

        pipeline_type = 'meta' if 'pipeline.enc0.0.t_blocks.0.gamma1' in state else 'linear'
        if pipeline_type == 'meta':
            num_blocks = [get_seq_len(state, f'pipeline.enc{index}.0.t_blocks') for index in range(4)]
            hidden_rate_shape = state['pipeline.enc0.0.t_blocks.0.ffn.key.weight'].shape
            channel_norm = 'pipeline.enc0.0.t_blocks.0.ffn.key_norm.weight' in state

        else:
            numbet = get_seq_len(state, 'pipeline.att')
            num_blocks = [get_seq_len(state, f'pipeline.att.{index}.t_blocks') for index in range(numbet)]
            hidden_rate_shape = state['pipeline.att.0.t_blocks.2.ffn.key.weight'].shape
            channel_norm = 'pipeline.att.0.t_blocks.0.ffn.key_norm.weight' in state
        hidden_rate = hidden_rate_shape[0] // hidden_rate_shape[1]
        if 'to_img.1.0.weight' in state:
            upsampler = 'n+c'
            scale = int(state['scale_factor'])
            end_index = get_seq_len(state, 'to_img.1') - 1
            out_channels = state[f'to_img.1.{end_index}.weight'].shape[0]
        elif 'to_img.init_pos' in state:
            upsampler = 'dys'
            out_channels = state['to_img.end_conv.weight'].shape[0]
            scale = math.isqrt(state['to_img.offset.weight'].shape[0] // 8)
        else:
            upsampler = 'ps'
            scale = math.isqrt(state['to_img.0.weight'].shape[0] // out_channels)
        model = FlexNet(
            inp_channels=inp_channels,
            out_channels=out_channels,
            scale=scale,
            dim=dim,
            num_blocks=num_blocks,
            window_size=window_size,
            hidden_rate=hidden_rate,
            channel_norm=channel_norm,
            pipeline_type=pipeline_type,
            upsampler=upsampler,
        )

        return self._enhance_model(model=model, in_channels=inp_channels, out_channels=out_channels, upscale=scale, name='FlexNet')



import math
from typing import Literal

import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange
from torch import Tensor, nn
from torch.nn.modules.module import T

SampleMods = Literal[
    'conv',
    'pixelshuffledirect',
    'pixelshuffle',
    'nearest+conv',
    'dysample',
    'transpose+conv',
    'lda',
    'pa_up',
]


class DySample(nn.Module):
    """Adapted from 'Learning to Upsample by Learning to Sample':
    https://arxiv.org/abs/2308.15085
    https://github.com/tiny-smart/dysample
    """

    def __init__(
        self,
        in_channels: int = 64,
        out_ch: int = 3,
        scale: int = 2,
        groups: int = 4,
        end_convolution: bool = True,
        end_kernel=1,
    ) -> None:
        super().__init__()

        if in_channels <= groups or in_channels % groups != 0:
            msg = 'Incorrect in_channels and groups values.'
            raise ValueError(msg)

        out_channels = 2 * groups * scale**2
        self.scale = scale
        self.groups = groups
        self.end_convolution = end_convolution
        if end_convolution:
            self.end_conv = nn.Conv2d(in_channels, out_ch, end_kernel, 1, end_kernel // 2)
        self.offset = nn.Conv2d(in_channels, out_channels, 1)
        self.scope = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        if self.training:
            nn.init.trunc_normal_(self.offset.weight, std=0.02)
            nn.init.constant_(self.scope.weight, val=0)

        self.register_buffer('init_pos', self._init_pos())

    def _init_pos(self) -> Tensor:
        h = torch.arange((-self.scale + 1) / 2, (self.scale - 1) / 2 + 1) / self.scale
        return torch.stack(torch.meshgrid([h, h], indexing='ij')).transpose(1, 2).repeat(1, self.groups, 1).reshape(1, -1, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        offset = self.offset(x) * self.scope(x).sigmoid() * 0.5 + self.init_pos
        B, _, H, W = offset.shape
        offset = offset.view(B, 2, -1, H, W)
        coords_h = torch.arange(H) + 0.5
        coords_w = torch.arange(W) + 0.5

        coords = (
            torch.stack(torch.meshgrid([coords_w, coords_h], indexing='ij'))
            .transpose(1, 2)
            .unsqueeze(1)
            .unsqueeze(0)
            .type(x.dtype)
            .to(x.device, non_blocking=True)
        )
        normalizer = torch.tensor([W, H], dtype=x.dtype, device=x.device, pin_memory=True).view(1, 2, 1, 1, 1)
        coords = 2 * (coords + offset) / normalizer - 1

        coords = (
            F.pixel_shuffle(coords.reshape(B, -1, H, W), self.scale)
            .view(B, 2, -1, self.scale * H, self.scale * W)
            .permute(0, 2, 3, 4, 1)
            .contiguous()
            .flatten(0, 1)
        )
        output = F.grid_sample(
            x.reshape(B * self.groups, -1, H, W),
            coords,
            mode='bilinear',
            align_corners=False,
            padding_mode='border',
        ).view(B, -1, self.scale * H, self.scale * W)

        if self.end_convolution:
            output = self.end_conv(output)

        return output


class LayerNorm(nn.Module):
    def __init__(self, dim: int = 64, eps: float = 1e-6) -> None:
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = eps
        self.dim = (dim,)

    def forward(self, x):
        if x.is_contiguous(memory_format=torch.channels_last):
            return F.layer_norm(x.permute(0, 2, 3, 1), self.dim, self.weight, self.bias, self.eps).permute(0, 3, 1, 2)
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight[:, None, None] * x + self.bias[:, None, None]


class LDA_AQU(nn.Module):
    def __init__(
        self,
        in_channels=48,
        reduction_factor=4,
        nh=1,
        scale_factor=2.0,
        k_e=3,
        k_u=3,
        n_groups=2,
        range_factor=11,
        rpb=True,
    ) -> None:
        super().__init__()
        self.k_u = k_u
        self.num_head = nh
        self.scale_factor = scale_factor
        self.n_groups = n_groups
        self.offset_range_factor = range_factor

        self.attn_dim = in_channels // (reduction_factor * self.num_head)
        self.scale = self.attn_dim**-0.5
        self.rpb = rpb
        self.hidden_dim = in_channels // reduction_factor
        self.proj_q = nn.Conv2d(in_channels, self.hidden_dim, kernel_size=1, stride=1, padding=0, bias=False)

        self.proj_k = nn.Conv2d(in_channels, self.hidden_dim, kernel_size=1, stride=1, padding=0, bias=False)

        self.group_channel = in_channels // (reduction_factor * self.n_groups)
        # print(self.group_channel)
        self.conv_offset = nn.Sequential(
            nn.Conv2d(
                self.group_channel,
                self.group_channel,
                3,
                1,
                1,
                groups=self.group_channel,
                bias=False,
            ),
            LayerNorm(self.group_channel),
            nn.SiLU(),
            nn.Conv2d(self.group_channel, 2 * k_u**2, k_e, 1, k_e // 2),
        )
        print(2 * k_u**2)
        self.layer_norm = LayerNorm(in_channels)

        self.pad = int((self.k_u - 1) / 2)
        base = np.arange(-self.pad, self.pad + 1).astype(np.float32)
        base_y = np.repeat(base, self.k_u)
        base_x = np.tile(base, self.k_u)
        base_offset = np.stack([base_y, base_x], axis=1).flatten()
        base_offset = torch.tensor(base_offset).view(1, -1, 1, 1)
        self.register_buffer('base_offset', base_offset, persistent=False)

        if self.rpb:
            self.relative_position_bias_table = nn.Parameter(torch.zeros(1, self.num_head, 1, self.k_u**2, self.hidden_dim // self.num_head))
            nn.init.trunc_normal_(self.relative_position_bias_table, std=0.02)

    def init_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform(m)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)
        nn.init.constant_(self.conv_offset[-1].weight, 0)
        nn.init.constant_(self.conv_offset[-1].bias, 0)

    def get_offset(self, offset, Hout, Wout):
        B, _, _, _ = offset.shape
        device = offset.device
        row_indices = torch.arange(Hout, device=device)
        col_indices = torch.arange(Wout, device=device)
        row_indices, col_indices = torch.meshgrid(row_indices, col_indices)
        index_tensor = torch.stack((row_indices, col_indices), dim=-1).view(1, Hout, Wout, 2)
        offset = rearrange(offset, 'b (kh kw d) h w -> b kh h kw w d', kh=self.k_u, kw=self.k_u)
        offset = offset + index_tensor.view(1, 1, Hout, 1, Wout, 2)
        offset = offset.contiguous().view(B, self.k_u * Hout, self.k_u * Wout, 2)

        offset[..., 0] = 2 * offset[..., 0] / (Hout - 1) - 1
        offset[..., 1] = 2 * offset[..., 1] / (Wout - 1) - 1
        offset = offset.flip(-1)
        return offset

    def extract_feats(self, x, offset, ks=3):
        out = nn.functional.grid_sample(x, offset, mode='bilinear', padding_mode='zeros', align_corners=True)
        out = rearrange(out, 'b c (ksh h) (ksw w) -> b (ksh ksw) c h w', ksh=ks, ksw=ks)
        return out

    def forward(self, x):
        B, C, H, W = x.shape
        out_H, out_W = int(H * self.scale_factor), int(W * self.scale_factor)
        v = x
        x = self.layer_norm(x)
        q = self.proj_q(x)
        k = self.proj_k(x)

        q = torch.nn.functional.interpolate(q, (out_H, out_W), mode='bilinear', align_corners=True)
        q_off = q.view(B * self.n_groups, -1, out_H, out_W)
        pred_offset = self.conv_offset(q_off)
        offset = pred_offset.tanh().mul(self.offset_range_factor) + self.base_offset.to(x.dtype)

        k = k.view(B * self.n_groups, self.hidden_dim // self.n_groups, H, W)
        v = v.view(B * self.n_groups, C // self.n_groups, H, W)
        offset = self.get_offset(offset, out_H, out_W)
        k = self.extract_feats(k, offset=offset)
        v = self.extract_feats(v, offset=offset)

        q = rearrange(q, 'b (nh c) h w -> b nh (h w) () c', nh=self.num_head)
        k = rearrange(k, '(b g) n c h w -> b (h w) n (g c)', g=self.n_groups)
        v = rearrange(v, '(b g) n c h w -> b (h w) n (g c)', g=self.n_groups)
        k = rearrange(k, 'b n1 n (nh c) -> b nh n1 n c', nh=self.num_head)
        v = rearrange(v, 'b n1 n (nh c) -> b nh n1 n c', nh=self.num_head)

        if self.rpb:
            k = k + self.relative_position_bias_table

        q = q * self.scale
        attn = q @ k.transpose(-1, -2)
        attn = attn.softmax(dim=-1)
        out = attn @ v

        out = rearrange(out, 'b nh (h w) t c -> b (nh c) (t h) w', h=out_H)
        return out


class PA(nn.Module):
    def __init__(self, dim) -> None:
        super().__init__()
        self.conv = nn.Sequential(nn.Conv2d(dim, dim, 1), nn.Sigmoid())

    def forward(self, x):
        return x.mul(self.conv(x))


class UniUpsampleV3(nn.Sequential):
    def __init__(
        self,
        upsample: SampleMods = 'pa_up',
        scale: int = 2,
        in_dim: int = 48,
        out_dim: int = 3,
        mid_dim: int = 48,
        group: int = 4,  # Only DySample
        dysample_end_kernel=1,  # needed only for compatibility with version 2
    ) -> None:
        m = []

        if scale == 1 or upsample == 'conv':
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'pixelshuffledirect':
            m.extend([nn.Conv2d(in_dim, out_dim * scale**2, 3, 1, 1), nn.PixelShuffle(scale)])
        elif upsample == 'pixelshuffle':
            m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            if (scale & (scale - 1)) == 0:  # scale = 2^n
                for _ in range(int(math.log2(scale))):
                    m.extend([nn.Conv2d(mid_dim, 4 * mid_dim, 3, 1, 1), nn.PixelShuffle(2)])
            elif scale == 3:
                m.extend([nn.Conv2d(mid_dim, 9 * mid_dim, 3, 1, 1), nn.PixelShuffle(3)])
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'nearest+conv':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        (
                            nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                            nn.Upsample(scale_factor=2),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        )
                    )
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            elif scale == 3:
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.Upsample(scale_factor=scale),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'dysample':
            if mid_dim != in_dim:
                m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            m.append(DySample(mid_dim, out_dim, scale, group, end_kernel=dysample_end_kernel))
            # m.append(nn.Conv2d(mid_dim, out_dim, dysample_end_kernel, 1, dysample_end_kernel//2)) # kernel 1 causes chromatic artifacts
        elif upsample == 'transpose+conv':
            if scale == 2:
                m.append(nn.ConvTranspose2d(in_dim, out_dim, 4, 2, 1))
            elif scale == 3:
                m.append(nn.ConvTranspose2d(in_dim, out_dim, 3, 3, 0))
            elif scale == 4:
                m.extend(
                    [
                        nn.ConvTranspose2d(in_dim, in_dim, 4, 2, 1),
                        nn.GELU(),
                        nn.ConvTranspose2d(in_dim, out_dim, 4, 2, 1),
                    ]
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2, 3, 4')
            m.append(nn.Conv2d(out_dim, out_dim, 3, 1, 1))
        elif upsample == 'lda':
            if mid_dim != in_dim:
                m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            m.append(LDA_AQU(mid_dim, scale_factor=scale))
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'pa_up':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        [
                            nn.Upsample(scale_factor=2),
                            nn.Conv2d(in_dim, mid_dim, 3, 1, 1),
                            PA(mid_dim),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                            nn.Conv2d(mid_dim, mid_dim, 3, 1, 1),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        ]
                    )
                    in_dim = mid_dim
            elif scale == 3:
                m.extend(
                    [
                        nn.Upsample(scale_factor=3),
                        nn.Conv2d(in_dim, mid_dim, 3, 1, 1),
                        PA(mid_dim),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(mid_dim, mid_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    ]
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        else:
            raise ValueError(f'An invalid Upsample was selected. Please choose one of {SampleMods}')
        super().__init__(*m)

        self.register_buffer(
            'MetaUpsample',
            torch.tensor(
                [
                    3,  # Block version, if you change something, please number from the end so that you can distinguish between authorized changes and third parties
                    list(SampleMods.__args__).index(upsample),  # UpSample method index
                    scale,
                    in_dim,
                    out_dim,
                    mid_dim,
                    group,
                ],
                dtype=torch.uint8,
            ),
        )


def pad_to_even(x, expand_all_sides=False):
    H, W = x.shape[-2:]

    pad_left = 0
    pad_right = W % 2
    pad_top = 0
    pad_bottom = H % 2

    if expand_all_sides:
        pad_left += 2
        pad_right += 2
        pad_top += 2
        pad_bottom += 2

    padded = F.pad(x, (pad_left, pad_right, pad_top, pad_bottom), mode='reflect')
    return padded, (pad_top, pad_bottom, pad_left, pad_right)


def unpad(x, pad):
    pad_top, pad_bottom, pad_left, pad_right = pad
    if pad_bottom:
        x = x[..., :-pad_bottom, :]
    if pad_top:
        x = x[..., pad_top:, :]
    if pad_right:
        x = x[..., :, :-pad_right]
    if pad_left:
        x = x[..., :, pad_left:]
    return x


class FourierUnit(nn.Module):
    def __init__(self, in_channels, out_channels, groups=1) -> None:
        super().__init__()
        self.groups = groups
        self.ln = LayerNorm(out_channels * 2)
        self.fdc = nn.Conv2d(
            in_channels=in_channels * 2,
            out_channels=out_channels * 2 * groups,
            kernel_size=1,
            groups=groups,
            bias=True,
        )
        self.weight = nn.Sequential(
            nn.Conv2d(in_channels=in_channels * 2, out_channels=groups, kernel_size=1),
            nn.Softmax(dim=1),
        )
        self.fpe = nn.Conv2d(
            in_channels=in_channels * 2,
            out_channels=in_channels * 2,
            kernel_size=3,
            padding=1,
            groups=in_channels * 2,
            bias=True,
        )
        self.pad = False

    def train(self: T, mode: bool = True) -> T:
        self.pad = not mode
        return super().train(mode)

    def forward(self, x):
        # сохраняем исходный dtype, чтобы вернуть
        x, pad = pad_to_even(x, self.pad)

        orig_dtype = x.dtype
        x = x.to(torch.float32)

        b, c, h, w = x.shape
        ffted = torch.fft.rfft2(x, norm='ortho')  # complex64
        real = torch.unsqueeze(torch.real(ffted), -1)  # float32
        imag = torch.unsqueeze(torch.imag(ffted), -1)  # float32
        ffted = rearrange(torch.cat((real, imag), -1), 'b c h w d -> b (c d) h w').to(orig_dtype)

        ffted = self.ln(ffted)  # float32
        ffted = self.fpe(ffted) + ffted  # float32

        dy_weight = self.weight(ffted)  # float32
        ffted = self.fdc(ffted).view(b, self.groups, 2 * c, h, -1)  # float32
        ffted = torch.einsum('ijkml,ijml->ikml', ffted, dy_weight)  # float32
        ffted = F.gelu(ffted)  # float32

        ffted = rearrange(ffted, 'b (c d) h w -> b c h w d', d=2).to(torch.float32)
        ffted = torch.view_as_complex(ffted.contiguous())  # complex64
        output = torch.fft.irfft2(ffted, s=(h, w), norm='ortho')  # float32

        return unpad(output.to(orig_dtype), pad)


class InceptionDWConv2d(nn.Module):
    """Inception depthweise convolution"""

    def __init__(
        self,
        in_channels,
        square_kernel_size=3,
        band_kernel_size=11,
        branch_ratio=0.125,
        shift=0,
        fft_mode=True,
    ) -> None:
        super().__init__()

        gc = int(in_channels * branch_ratio)  # channel numbers of a convolution branch
        convs = [
            nn.Identity(),
            nn.Conv2d(gc, gc, square_kernel_size, padding=square_kernel_size // 2, groups=gc),
            nn.Conv2d(
                gc,
                gc,
                kernel_size=(1, band_kernel_size),
                padding=(0, band_kernel_size // 2),
                groups=gc,
            ),
            nn.Conv2d(
                gc,
                gc,
                kernel_size=(band_kernel_size, 1),
                padding=(band_kernel_size // 2, 0),
                groups=gc,
            ),
            FourierUnit(gc, gc) if fft_mode else nn.Identity(),
        ]
        self.pconv = convs[shift % 5]
        self.dwconv_hw = convs[(shift + 1) % 5]
        self.dwconv_w = convs[(shift + 2) % 5]
        self.dwconv_h = convs[(shift + 3) % 5]
        self.fsas = convs[(shift + 4) % 5]
        indexs = [in_channels - 4 * gc, gc, gc, gc, gc]
        self.split_indexes = (
            indexs[shift % 5],
            indexs[(shift + 1) % 5],
            indexs[(shift + 2) % 5],
            indexs[(shift + 3) % 5],
            indexs[(shift + 4) % 5],
        )

    def forward(self, x):
        x_id, x_hw, x_w, x_h, fsas = torch.split(x, self.split_indexes, dim=1)
        return torch.cat(
            (
                self.pconv(x_id),
                self.dwconv_hw(x_hw),
                self.dwconv_w(x_w),
                self.dwconv_h(x_h),
                self.fsas(fsas),
            ),
            dim=1,
        )


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(
        self,
        dim: int = 64,
        expansion_ratio: float = 8 / 3,
        fft_mode=True,
        shift: int = 0,
    ) -> None:
        super().__init__()
        self.norm = LayerNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Conv2d(dim, hidden * 2, 3, 1, 1)

        self.act = nn.Mish()
        conv_channels = dim
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]
        self.conv = InceptionDWConv2d(conv_channels, shift=shift, fft_mode=fft_mode)
        self.fc2 = nn.Conv2d(hidden, dim, 3, 1, 1)
        self.gamma = nn.Parameter(torch.ones([1, dim, 1, 1]), requires_grad=True)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        shortcut = x

        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.conv(c)
        x = self.act(self.fc2(self.act(g) * torch.cat((i, c), dim=1)))

        return x * self.gamma + shortcut


class PadPixelUnshuffle(nn.Module):
    r"""PadPixelUnshuffle."""

    def __init__(self, down: int) -> None:
        super().__init__()
        self.ps = down

    def forward(self, x):
        b, c, h, w = x.shape
        mod_pad_h = (self.ps - h % self.ps) % self.ps
        mod_pad_w = (self.ps - w % self.ps) % self.ps
        return F.pixel_unshuffle(F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect'), self.ps)


class GFISR(nn.Module):
    def __init__(
        self,
        in_nc=3,
        dim=48,
        expansion_ratio=8 / 3,
        fft_mode=True,
        scale=4,
        out_nc=3,
        upsampler: SampleMods = 'pixelshuffledirect',
        mid_dim=32,
        pixel_unshuffle=True,
        n_blocks=24,
        **kwargs,
    ) -> None:
        super().__init__()
        self.scale = scale
        if pixel_unshuffle and scale in [1, 2]:
            down = 4 // scale
            self.in_to_dim = nn.Sequential(PadPixelUnshuffle(down), nn.Conv2d(in_nc * down * down, dim, 3, 1, 1))
            scale = 4
        else:
            self.in_to_dim = nn.Conv2d(in_nc, dim, 3, 1, 1)
        self.net = nn.Sequential(*[GatedCNNBlock(dim, shift=i, expansion_ratio=expansion_ratio, fft_mode=fft_mode) for i in range(n_blocks)])
        self.dim_to_out = UniUpsampleV3(upsampler, scale, dim, out_nc, mid_dim, dysample_end_kernel=3)

    def load_state_dict(self, state_dict, *args, **kwargs):
        state_dict['dim_to_out.MetaUpsample'] = self.dim_to_out.MetaUpsample
        print(state_dict['dim_to_out.MetaUpsample'])
        return super().load_state_dict(state_dict, *args, **kwargs)

    def forward(self, x):
        b, c, h, w = x.shape
        x = self.in_to_dim(x)
        x = self.net(x) + x
        return self.dim_to_out(x)[:, :, : h * self.scale, : w * self.scale]



from typing import Mapping

from .arch import SPAN
import torch

from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import pixelshuffle_scale


class SPANArch(Architecture[SPAN]):
    def __init__(self):
        super().__init__(
            uid='SPAN',
            detect=KeyCondition.has_all(
                'conv_1.sk.weight',
                'block_1.c1_r.sk.weight',
                'block_1.c1_r.eval_conv.weight',
                'block_1.c3_r.eval_conv.weight',
                'conv_cat.weight',
                'conv_2.sk.weight',
                'conv_2.eval_conv.weight',
                'upsampler.0.weight',
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        norm = True
        img_range = 255.0  # cannot be deduced from state_dict
        rgb_mean = (0.4488, 0.4371, 0.4040)  # cannot be deduced from state_dict

        num_in_ch = state_dict['conv_1.sk.weight'].shape[1]
        feature_channels = state_dict['conv_1.sk.weight'].shape[0]
        num_out_ch = num_in_ch
        # pixelshuffel shenanigans
        upscale = pixelshuffle_scale(
            state_dict['upsampler.0.weight'].shape[0],
            num_in_ch,
        )

        # norm
        if 'no_norm' in state_dict:
            norm = False
            state_dict['no_norm'] = torch.zeros(1)

        model = SPAN(
            num_in_ch=num_in_ch,
            num_out_ch=num_out_ch,
            feature_channels=feature_channels,
            upscale=upscale,
            norm=norm,
            img_range=img_range,
            rgb_mean=rgb_mean,
        )

        return self._enhance_model(model=model, in_channels=num_in_ch, out_channels=num_out_ch, upscale=upscale, name='SPAN')



import torch
import torch.nn.functional as F
from torch import nn
from torch.nn.init import trunc_normal_


class CSELayer(nn.Module):
    def __init__(self, num_channels: int = 48, reduction_ratio: int = 2):
        super(CSELayer, self).__init__()
        num_channels_reduced = num_channels // reduction_ratio
        self.squeezing = nn.Sequential(
            nn.Conv2d(num_channels, num_channels_reduced, 1, 1),
            nn.ReLU(True),
            nn.Conv2d(num_channels_reduced, num_channels, 1, 1),
            nn.Hardsigmoid(True),
        )

    def forward(self, input_tensor):
        squeeze_tensor = torch.mean(input_tensor, dim=[2, 3], keepdim=True)
        output_tensor = input_tensor * self.squeezing(squeeze_tensor)
        return output_tensor


# channels_first
class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6) -> None:
        super(RMSNorm, self).__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(dim))
        self.offset = nn.Parameter(torch.zeros(dim))

    def forward(self, x):
        norm_x = x.norm(2, dim=1, keepdim=True)
        d_x = x.size(1)
        rms_x = norm_x * (d_x ** (-1.0 / 2))
        x_normed = x / (rms_x + self.eps)
        return self.scale[..., None, None] * x_normed + self.offset[..., None, None]


class Conv3XC(nn.Module):
    def __init__(self, c_in: int, c_out: int, gain: int = 2, s: int = 1, bias: bool = True) -> None:
        super().__init__()
        self.weight_concat = None
        self.bias_concat = None
        self.update_params_flag = False
        self.stride = s

        self.sk = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=1,
            padding=0,
            stride=s,
            bias=bias,
        )
        self.conv = nn.Sequential(
            nn.Conv2d(
                in_channels=c_in,
                out_channels=c_in * gain,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_in * gain,
                out_channels=c_out * gain,
                kernel_size=3,
                stride=s,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_out * gain,
                out_channels=c_out,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
        )
        self.eval_conv = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=3,
            padding=1,
            stride=s,
            bias=bias,
        )

    def update_params(self) -> None:
        w1 = self.conv[0].weight.data.clone().detach()
        b1 = self.conv[0].bias.data.clone().detach()
        w2 = self.conv[1].weight.data.clone().detach()
        b2 = self.conv[1].bias.data.clone().detach()
        w3 = self.conv[2].weight.data.clone().detach()
        b3 = self.conv[2].bias.data.clone().detach()

        w = F.conv2d(w1.flip(2, 3).permute(1, 0, 2, 3), w2, padding=2, stride=1).flip(2, 3).permute(1, 0, 2, 3)
        b = (w2 * b1.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b2

        self.weight_concat = F.conv2d(w.flip(2, 3).permute(1, 0, 2, 3), w3, padding=0, stride=1).flip(2, 3).permute(1, 0, 2, 3)
        self.bias_concat = (w3 * b.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b3

        sk_w = self.sk.weight.data.clone().detach()
        sk_b = self.sk.bias.data.clone().detach()  # type: ignore
        target_kernel_size = 3

        H_pixels_to_pad = (target_kernel_size - 1) // 2  # noqa: N806
        W_pixels_to_pad = (target_kernel_size - 1) // 2  # noqa: N806
        sk_w = F.pad(sk_w, [H_pixels_to_pad, H_pixels_to_pad, W_pixels_to_pad, W_pixels_to_pad])

        self.weight_concat = self.weight_concat + sk_w
        self.bias_concat = self.bias_concat + sk_b

        self.eval_conv.weight.data = self.weight_concat
        self.eval_conv.bias.data = self.bias_concat  # type: ignore

    def forward(self, x):  # noqa: ANN201, ANN001
        x_pad = F.pad(x, (1, 1, 1, 1), 'constant', 0)
        out = self.conv(x_pad) + self.sk(x)
        return out


class SeqConv3x3(nn.Module):
    def __init__(self, inp_planes, out_planes, depth_multiplier):
        super(SeqConv3x3, self).__init__()
        self.inp_planes = inp_planes
        self.out_planes = out_planes
        self.mid_planes = int(out_planes * depth_multiplier)
        conv0 = torch.nn.Conv2d(self.inp_planes, self.mid_planes, kernel_size=1, padding=0)
        self.k0 = conv0.weight
        self.b0 = conv0.bias

        conv1 = torch.nn.Conv2d(self.mid_planes, self.out_planes, kernel_size=3)
        self.k1 = conv1.weight
        self.b1 = conv1.bias

    def forward(self, x):
        # conv-1x1
        y0 = F.conv2d(input=x, weight=self.k0, bias=self.b0, stride=1)
        # explicitly padding with bias
        y0 = F.pad(y0, (1, 1, 1, 1), 'constant', 0)
        b0_pad = self.b0.view(1, -1, 1, 1)
        y0[:, :, 0:1, :] = b0_pad
        y0[:, :, -1:, :] = b0_pad
        y0[:, :, :, 0:1] = b0_pad
        y0[:, :, :, -1:] = b0_pad
        # conv-3x3
        return F.conv2d(input=y0, weight=self.k1, bias=self.b1, stride=1)

    def rep_params(self):
        device = self.k0.get_device()
        if device < 0:
            device = None
        # re-param conv kernel
        RK = F.conv2d(input=self.k1, weight=self.k0.permute(1, 0, 2, 3))
        # re-param conv bias
        RB = torch.ones(1, self.mid_planes, 3, 3, device=device) * self.b0.view(1, -1, 1, 1)
        RB = (
            F.conv2d(input=RB, weight=self.k1).view(
                -1,
            )
            + self.b1
        )
        return RK, RB


class RepConv(nn.Module):
    def __init__(self, in_dim=3, out_dim=32):
        super().__init__()
        self.conv1 = SeqConv3x3(in_dim, out_dim, 2)
        self.conv2 = nn.Conv2d(in_dim, out_dim, 3, 1, 1)
        self.conv3 = Conv3XC(in_dim, out_dim)
        self.conv_3x3_rep = nn.Conv2d(in_dim, out_dim, 3, 1, 1)
        self.alpha = nn.Parameter(torch.randn(3), requires_grad=True)
        self.forward_module = self.train_forward

        nn.init.constant_(self.alpha, 1.0)

    def fuse(self):
        conv1_w, conv1_b = self.conv1.rep_params()
        conv2_w, conv2_b = self.conv2.weight, self.conv2.bias
        self.conv3.update_params()
        conv3_w, conv3_b = self.conv3.eval_conv.weight, self.conv3.eval_conv.bias
        device = self.conv_3x3_rep.weight.device
        sum_weight = (self.alpha[0] * conv1_w + self.alpha[1] * conv2_w + self.alpha[2] * conv3_w).to(device)
        sum_bias = (self.alpha[0] * conv1_b + self.alpha[1] * conv2_b + self.alpha[2] * conv3_b).to(device)
        self.conv_3x3_rep.weight = nn.Parameter(sum_weight)
        self.conv_3x3_rep.bias = nn.Parameter(sum_bias)

    def train_forward(self, x):
        x1 = self.conv1(x)
        x2 = self.conv2(x)
        x3 = self.conv3(x)
        return self.alpha[0] * x1 + self.alpha[1] * x2 + self.alpha[2] * x3

    def train(self, mode: bool = True):
        super().train(mode)
        if not mode:
            self.fuse()
        return self

    def forward(self, x):
        if self.training:
            return self.train_forward(x)
        else:
            return self.conv_3x3_rep(x)


class OmniShift(nn.Module):
    def __init__(self, dim: int = 48) -> None:
        super().__init__()
        # Define the layers for training
        self.conv1x1 = nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=1, groups=dim, bias=True)
        self.conv3x3 = nn.Conv2d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=3,
            padding=1,
            groups=dim,
            bias=True,
        )
        self.conv5x5 = nn.Conv2d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=5,
            padding=2,
            groups=dim,
            bias=True,
        )
        self.alpha1 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)
        self.alpha2 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)
        self.alpha3 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)
        self.alpha4 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)

        # Define the layers for testing
        self.conv5x5_reparam = nn.Conv2d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=5,
            padding=2,
            groups=dim,
            bias=True,
        )

    def forward_train(self, x):
        out1x1 = self.conv1x1(x)
        out3x3 = self.conv3x3(x)
        out5x5 = self.conv5x5(x)

        out = self.alpha1 * x + self.alpha2 * out1x1 + self.alpha3 * out3x3 + self.alpha4 * out5x5
        return out

    def reparam_5x5(self) -> None:
        # Combine the parameters of conv1x1, conv3x3, and conv5x5 to form a single 5x5 depth-wise convolution

        padded_weight_1x1 = F.pad(self.conv1x1.weight, (2, 2, 2, 2))
        padded_weight_3x3 = F.pad(self.conv3x3.weight, (1, 1, 1, 1))

        identity_weight = F.pad(torch.ones_like(self.conv1x1.weight), (2, 2, 2, 2))
        combined_weight = (
            self.alpha1.transpose(0, 1) * identity_weight
            + self.alpha2.transpose(0, 1) * padded_weight_1x1
            + self.alpha3.transpose(0, 1) * padded_weight_3x3
            + self.alpha4.transpose(0, 1) * self.conv5x5.weight
        )

        combined_bias = (
            self.alpha2.squeeze() * self.conv1x1.bias + self.alpha3.squeeze() * self.conv3x3.bias + self.alpha4.squeeze() * self.conv5x5.bias
        )

        device = self.conv5x5_reparam.weight.device

        combined_weight = combined_weight.to(device)
        combined_bias = combined_bias.to(device)

        self.conv5x5_reparam.weight = nn.Parameter(combined_weight)
        self.conv5x5_reparam.bias = nn.Parameter(combined_bias)

    def train(self, mode: bool = True):
        super().train(mode)
        if not mode:
            self.reparam_5x5()

    def forward(self, x):
        if self.training:
            out = self.forward_train(x)
        else:
            out = self.conv5x5_reparam(x)
        return out


class ParPixelUnshuffle(nn.Module):
    def __init__(self, in_dim, out_dim, down):
        super().__init__()
        self.pu = nn.PixelUnshuffle(down)
        self.poll = nn.Sequential(nn.MaxPool2d(kernel_size=down, stride=down), RepConv(in_dim, out_dim))

    def forward(self, x):
        return self.pu(x) + self.poll(x)


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(
        self,
        dim: int = 64,
        expansion_ratio: float = 8 / 3,
        conv_ratio: float = 1.0,
        dccm: bool = True,
        se: bool = False,
    ) -> None:
        super().__init__()
        self.norm = RMSNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = RepConv(dim, hidden * 2)
        self.act = nn.Mish()
        conv_channels = int(conv_ratio * dim)
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]
        self.conv = nn.Sequential(
            ParPixelUnshuffle(dim, dim * 4, 2),
            OmniShift(dim * 4),
            CSELayer(dim * 4) if se else nn.Identity(),
            nn.PixelShuffle(2),
        )  # InceptionDWConv2d(dim*4)
        self.fc2 = RepConv(hidden, dim) if dccm else nn.Conv2d(hidden, dim, 1, 1)

    def forward(self, x):
        shortcut = x
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.conv(c)
        x = self.act(self.fc2(self.act(g) * torch.cat((i, c), dim=1)))
        return x + shortcut


class RTMoSR(nn.Module):
    def __init__(
        self,
        scale: int = 2,
        dim: int = 32,
        ffn_expansion: float = 2,
        n_blocks: int = 2,
        unshuffle_mod: bool = False,
        dccm: bool = True,
        se: bool = True,
    ):
        super().__init__()
        self.scale = scale
        unshuffle = 0
        if scale < 4 and unshuffle_mod:
            if scale == 3:
                raise ValueError('Unshuffle_mod does not support 3x')
            unshuffle = 4 // scale
            scale = 4
        self.pad = unshuffle if unshuffle > 0 else 1
        self.pad *= 2
        self.to_feat = RepConv(3, dim) if not unshuffle else nn.Sequential(nn.PixelUnshuffle(unshuffle), RepConv(3 * unshuffle * unshuffle, dim))
        self.body = nn.Sequential(*[GatedCNNBlock(dim, ffn_expansion, dccm=dccm, se=se) for _ in range(n_blocks)])
        self.to_img = nn.Sequential(
            RepConv(dim, 3 * scale**2),
            nn.PixelShuffle(scale),
        )
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def check_img_size(self, x, resolution: tuple[int, int]):
        scaled_size = self.pad
        mod_pad_h = (scaled_size - resolution[0] % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - resolution[1] % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, x):
        b, c, h, w = x.shape
        out = self.check_img_size(x, (h, w))
        out = self.to_feat(out)
        out = self.body(out)
        return self.to_img(out)[:, :, : h * self.scale, : w * self.scale] + F.interpolate(x, scale_factor=self.scale)



"""
cache_mode:
0:使用cache缓存必要参数
1:使用cache缓存必要参数,对cache进行8bit量化节省显存,带来小许延时增长
2:不使用cache,耗时约为mode0的2倍,但是显存不受输入图像分辨率限制,tile_mode填得够大,1.5G显存可超任意比例
"""

from __future__ import annotations

import torch
from torch import Tensor
from torch import nn as nn
from torch.nn import functional as F


def q(inp, cache_mode):
    maxx = inp.max()
    minn = inp.min()
    delta = maxx - minn
    if cache_mode == 2:
        return (
            ((inp - minn) / delta * 255).round().byte().cpu(),
            delta,
            minn,
            inp.device,
        )  # 大概3倍延时#太慢了，屏蔽该模式
    elif cache_mode == 1:
        return (
            ((inp - minn) / delta * 255).round().byte(),
            delta,
            minn,
            inp.device,
        )  # 不用CPU转移


def dq(inp, if_half: bool, cache_mode, delta, minn, device):
    if cache_mode == 2:
        if if_half:
            return inp.to(device).half() / 255 * delta + minn
        else:
            return inp.to(device).float() / 255 * delta + minn
    elif cache_mode == 1:
        if if_half:
            return inp.half() / 255 * delta + minn  # 不用CPU转移
        else:
            return inp.float() / 255 * delta + minn
    else:
        raise ValueError('cache_mode config error')


class SEBlock(nn.Module):
    def __init__(self, in_channels: int, reduction=8, bias=False):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, in_channels // reduction, 1, 1, 0, bias=bias)
        self.conv2 = nn.Conv2d(in_channels // reduction, in_channels, 1, 1, 0, bias=bias)

    def forward(self, x):
        if 'Half' in x.type():  # torch.HalfTensor/torch.cuda.HalfTensor
            x0 = torch.mean(x.float(), dim=(2, 3), keepdim=True).half()
        else:
            x0 = torch.mean(x, dim=(2, 3), keepdim=True)
        x0 = self.conv1(x0)
        x0 = F.relu(x0, inplace=True)
        x0 = self.conv2(x0)
        x0 = torch.sigmoid(x0)
        x = torch.mul(x, x0)
        return x

    def forward_mean(self, x, x0):
        x0 = self.conv1(x0)
        x0 = F.relu(x0, inplace=True)
        x0 = self.conv2(x0)
        x0 = torch.sigmoid(x0)
        x = torch.mul(x, x0)
        return x


class UNetConv(nn.Module):
    def __init__(self, in_channels: int, mid_channels: int, out_channels: int, se: bool):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, 3, 1, 0),
            nn.LeakyReLU(0.1, inplace=True),
            nn.Conv2d(mid_channels, out_channels, 3, 1, 0),
            nn.LeakyReLU(0.1, inplace=True),
        )
        if se:
            self.seblock = SEBlock(out_channels, reduction=8, bias=True)
        else:
            self.seblock = None

    def forward(self, x):
        z = self.conv(x)
        if self.seblock is not None:
            z = self.seblock(z)
        return z


class UNet1(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, deconv: bool):
        super().__init__()
        self.conv1 = UNetConv(in_channels, 32, 64, se=False)
        self.conv1_down = nn.Conv2d(64, 64, 2, 2, 0)
        self.conv2 = UNetConv(64, 128, 64, se=True)
        self.conv2_up = nn.ConvTranspose2d(64, 64, 2, 2, 0)
        self.conv3 = nn.Conv2d(64, 64, 3, 1, 0)

        if deconv:
            self.conv_bottom = nn.ConvTranspose2d(64, out_channels, 4, 2, 3)
        else:
            self.conv_bottom = nn.Conv2d(64, out_channels, 3, 1, 0)

        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:  # type: ignore
                    nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x1 = self.conv1(x)
        x2 = self.conv1_down(x1)
        x1 = F.pad(x1, (-4, -4, -4, -4))
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x2 = self.conv2(x2)
        x2 = self.conv2_up(x2)
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x3 = self.conv3(x1 + x2)
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        z = self.conv_bottom(x3)
        return z

    def forward_a(self, x):
        x1 = self.conv1(x)
        x2 = self.conv1_down(x1)
        x1 = F.pad(x1, (-4, -4, -4, -4))
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x2 = self.conv2.conv(x2)
        return x1, x2

    def forward_b(self, x1, x2):
        x2 = self.conv2_up(x2)
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x3 = self.conv3(x1 + x2)
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        z = self.conv_bottom(x3)
        return z


class UNet1x3(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, deconv: bool):
        super().__init__()
        self.conv1 = UNetConv(in_channels, 32, 64, se=False)
        self.conv1_down = nn.Conv2d(64, 64, 2, 2, 0)
        self.conv2 = UNetConv(64, 128, 64, se=True)
        self.conv2_up = nn.ConvTranspose2d(64, 64, 2, 2, 0)
        self.conv3 = nn.Conv2d(64, 64, 3, 1, 0)

        if deconv:
            self.conv_bottom = nn.ConvTranspose2d(64, out_channels, 5, 3, 2)
        else:
            self.conv_bottom = nn.Conv2d(64, out_channels, 3, 1, 0)

        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:  # type: ignore
                    nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x1 = self.conv1(x)
        x2 = self.conv1_down(x1)
        x1 = F.pad(x1, (-4, -4, -4, -4))
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x2 = self.conv2(x2)
        x2 = self.conv2_up(x2)
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x3 = self.conv3(x1 + x2)
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        z = self.conv_bottom(x3)
        return z

    def forward_a(self, x):
        x1 = self.conv1(x)
        x2 = self.conv1_down(x1)
        x1 = F.pad(x1, (-4, -4, -4, -4))
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x2 = self.conv2.conv(x2)
        return x1, x2

    def forward_b(self, x1, x2):
        x2 = self.conv2_up(x2)
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x3 = self.conv3(x1 + x2)
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        z = self.conv_bottom(x3)
        return z


class UNet2(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, deconv: bool):
        super().__init__()

        self.conv1 = UNetConv(in_channels, 32, 64, se=False)
        self.conv1_down = nn.Conv2d(64, 64, 2, 2, 0)
        self.conv2 = UNetConv(64, 64, 128, se=True)
        self.conv2_down = nn.Conv2d(128, 128, 2, 2, 0)
        self.conv3 = UNetConv(128, 256, 128, se=True)
        self.conv3_up = nn.ConvTranspose2d(128, 128, 2, 2, 0)
        self.conv4 = UNetConv(128, 64, 64, se=True)
        self.conv4_up = nn.ConvTranspose2d(64, 64, 2, 2, 0)
        self.conv5 = nn.Conv2d(64, 64, 3, 1, 0)

        if deconv:
            self.conv_bottom = nn.ConvTranspose2d(64, out_channels, 4, 2, 3)
        else:
            self.conv_bottom = nn.Conv2d(64, out_channels, 3, 1, 0)

        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:  # type: ignore
                    nn.init.constant_(m.bias, 0)

    def forward(self, x, alpha: float = 1):
        x1 = self.conv1(x)
        x2 = self.conv1_down(x1)
        x1 = F.pad(x1, (-16, -16, -16, -16))
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x2 = self.conv2(x2)
        x3 = self.conv2_down(x2)
        x2 = F.pad(x2, (-4, -4, -4, -4))
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        x3 = self.conv3(x3)
        x3 = self.conv3_up(x3)
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        x4 = self.conv4(x2 + x3)
        x4 *= alpha
        x4 = self.conv4_up(x4)
        x4 = F.leaky_relu(x4, 0.1, inplace=True)
        x5 = self.conv5(x1 + x4)
        x5 = F.leaky_relu(x5, 0.1, inplace=True)
        z = self.conv_bottom(x5)
        return z

    def forward_a(self, x):  # conv234结尾有se
        x1 = self.conv1(x)
        x2 = self.conv1_down(x1)
        x1 = F.pad(x1, (-16, -16, -16, -16))
        x2 = F.leaky_relu(x2, 0.1, inplace=True)
        x2 = self.conv2.conv(x2)
        return x1, x2

    def forward_b(self, x2):  # conv234结尾有se
        x3 = self.conv2_down(x2)
        x2 = F.pad(x2, (-4, -4, -4, -4))
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        x3 = self.conv3.conv(x3)
        return x2, x3

    def forward_c(self, x2, x3):  # conv234结尾有se
        x3 = self.conv3_up(x3)
        x3 = F.leaky_relu(x3, 0.1, inplace=True)
        x4 = self.conv4.conv(x2 + x3)
        return x4

    def forward_d(self, x1, x4):  # conv234结尾有se
        x4 = self.conv4_up(x4)
        x4 = F.leaky_relu(x4, 0.1, inplace=True)
        x5 = self.conv5(x1 + x4)
        x5 = F.leaky_relu(x5, 0.1, inplace=True)

        z = self.conv_bottom(x5)
        return z


class UpCunet2x(nn.Module):
    hyperparameters = {}

    def __init__(self, *, in_channels=3, out_channels=3, pro: bool = False):
        super().__init__()
        self.pro: Tensor | None
        if pro:
            self.register_buffer('pro', torch.zeros(1))
        else:
            self.pro = None

        self.unet1 = UNet1(in_channels, out_channels, deconv=True)
        self.unet2 = UNet2(in_channels, out_channels, deconv=False)

    @property
    def is_pro(self):
        return self.pro is not None

    def forward(self, x: Tensor, alpha: float = 1):
        _, _, h0, w0 = x.shape

        if self.is_pro:
            # pro expects a different input range
            x = x * 0.7 + 0.15

        ph = ((h0 - 1) // 2 + 1) * 2
        pw = ((w0 - 1) // 2 + 1) * 2
        x = F.pad(x, (18, 18 + pw - w0, 18, 18 + ph - h0), 'reflect')  # 需要保证被2整除
        x = self.unet1.forward(x)
        x0 = self.unet2.forward(x, alpha)
        x = F.pad(x, (-20, -20, -20, -20))
        x = torch.add(x0, x)
        if w0 != pw or h0 != ph:
            x = x[:, :, : h0 * 2, : w0 * 2]

        if self.is_pro:
            x = (x - 0.15) / 0.7

        return x


class UpCunet3x(nn.Module):
    hyperparameters = {}

    def __init__(self, *, in_channels=3, out_channels=3, pro: bool = False):
        super().__init__()
        self.pro: Tensor | None
        if pro:
            self.register_buffer('pro', torch.zeros(1))
        else:
            self.pro = None

        self.unet1 = UNet1x3(in_channels, out_channels, deconv=True)
        self.unet2 = UNet2(in_channels, out_channels, deconv=False)

    @property
    def is_pro(self):
        return self.pro is not None

    def forward(self, x: Tensor, alpha: float = 1):
        _, _, h0, w0 = x.shape

        if self.is_pro:
            # pro expects a different input range
            x = x * 0.7 + 0.15

        ph = ((h0 - 1) // 4 + 1) * 4
        pw = ((w0 - 1) // 4 + 1) * 4
        x = F.pad(x, (14, 14 + pw - w0, 14, 14 + ph - h0), 'reflect')  # 需要保证被2整除
        x = self.unet1.forward(x)
        x0 = self.unet2.forward(x, alpha)
        x = F.pad(x, (-20, -20, -20, -20))
        x = torch.add(x0, x)
        if w0 != pw or h0 != ph:
            x = x[:, :, : h0 * 3, : w0 * 3]

        if self.is_pro:
            x = (x - 0.15) / 0.7

        return x


class UpCunet4x(nn.Module):
    hyperparameters = {}

    def __init__(self, *, in_channels=3, out_channels=3, pro: bool = False):
        super().__init__()
        self.pro: Tensor | None
        if pro:
            self.register_buffer('pro', torch.zeros(1))
        else:
            self.pro = None

        self.unet1 = UNet1(in_channels, 64, deconv=True)
        self.unet2 = UNet2(64, 64, deconv=False)
        self.ps = nn.PixelShuffle(2)
        self.conv_final = nn.Conv2d(64, 12, 3, 1, padding=0, bias=True)

    @property
    def is_pro(self):
        return self.pro is not None

    def forward(self, x: Tensor, alpha: float = 1):
        _, _, h0, w0 = x.shape

        if self.is_pro:
            # pro expects a different input range
            x = x * 0.7 + 0.15

        x00 = x

        ph = ((h0 - 1) // 2 + 1) * 2
        pw = ((w0 - 1) // 2 + 1) * 2
        x = F.pad(x, (19, 19 + pw - w0, 19, 19 + ph - h0), 'reflect')  # 需要保证被2整除
        x = self.unet1.forward(x)
        x0 = self.unet2.forward(x, alpha)
        x1 = F.pad(x, (-20, -20, -20, -20))
        x = torch.add(x0, x1)
        x = self.conv_final(x)
        x = F.pad(x, (-1, -1, -1, -1))
        x = self.ps(x)
        if w0 != pw or h0 != ph:
            x = x[:, :, : h0 * 4, : w0 * 4]
        x += F.interpolate(x00, scale_factor=4, mode='nearest')

        if self.is_pro:
            x = (x - 0.15) / 0.7

        return x


class UpCunet2x_fast(nn.Module):
    hyperparameters = {}

    def __init__(self, *, in_channels=3, out_channels=3):
        super().__init__()
        self.unet1 = UNet1(12, 64, deconv=True)
        self.unet2 = UNet2(64, 64, deconv=False)
        self.ps = nn.PixelShuffle(2)
        self.conv_final = nn.Conv2d(64, 12, 3, 1, padding=0, bias=True)
        self.inv = nn.PixelUnshuffle(2)

    def forward(self, x: Tensor):
        _, _, h0, w0 = x.shape
        x00 = x
        ph = ((h0 - 1) // 2 + 1) * 2
        pw = ((w0 - 1) // 2 + 1) * 2
        x = F.pad(x, (38, 38 + pw - w0, 38, 38 + ph - h0), 'reflect')  # 需要保证被2整除
        x = self.inv(x)  # +18
        x = self.unet1.forward(x)
        x0 = self.unet2.forward(x)
        x1 = F.pad(x, (-20, -20, -20, -20))
        x = torch.add(x0, x1)
        x = self.conv_final(x)
        x = F.pad(x, (-1, -1, -1, -1))
        x = self.ps(x)
        if w0 != pw or h0 != ph:
            x = x[:, :, : h0 * 2, : w0 * 2]
        x += F.interpolate(x00, scale_factor=2, mode='nearest')
        return x



from typing import Mapping, Union, Literal

from .arch import UpCunet2x, UpCunet2x_fast, UpCunet3x, UpCunet4x
import torch

from ...factory import Architecture, KeyCondition

_CUGAN = Union[UpCunet2x, UpCunet3x, UpCunet4x, UpCunet2x_fast]


class CUGANArch(Architecture[_CUGAN]):
    def __init__(self):
        super().__init__(
            uid='CuGAN',
            detect=KeyCondition.has_all(
                'unet1.conv1.conv.0.weight',
                'unet1.conv1.conv.2.weight',
                'unet1.conv1_down.weight',
                'unet1.conv2.conv.0.weight',
                'unet1.conv2.conv.2.weight',
                'unet1.conv2.seblock.conv1.weight',
                'unet1.conv2_up.weight',
                'unet1.conv_bottom.weight',
                'unet2.conv1.conv.0.weight',
                'unet2.conv1_down.weight',
                'unet2.conv2.conv.0.weight',
                'unet2.conv2.seblock.conv1.weight',
                'unet2.conv3.conv.0.weight',
                'unet2.conv3.seblock.conv1.weight',
                'unet2.conv3_up.weight',
                'unet2.conv4.conv.0.weight',
                'unet2.conv4_up.weight',
                'unet2.conv5.weight',
                'unet2.conv_bottom.weight',
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        scale: Literal[2, 3, 4]
        in_channels: int
        out_channels: int
        pro: bool = False

        tags: list[str] = []
        if 'pro' in state_dict:
            pro = True
            tags.append('pro')
            state_dict['pro'] = torch.zeros(1)

        in_channels = state_dict['unet1.conv1.conv.0.weight'].shape[1]

        if 'conv_final.weight' in state_dict and in_channels == 12:
            # UpCunet2x_fast
            scale = 2
            in_channels = 3  # hard coded in UpCunet2x_fast
            out_channels = 3  # hard coded in UpCunet2x_fast
            model = UpCunet2x_fast(in_channels=in_channels, out_channels=out_channels)
            tags.append('fast')
        elif 'conv_final.weight' in state_dict:
            # UpCunet4x
            scale = 4
            out_channels = 3  # hard coded in UpCunet4x
            model = UpCunet4x(in_channels=in_channels, out_channels=out_channels, pro=pro)
        elif state_dict['unet1.conv_bottom.weight'].shape[2] == 5:
            # UpCunet3x
            scale = 3
            out_channels = state_dict['unet2.conv_bottom.weight'].shape[0]
            model = UpCunet3x(in_channels=in_channels, out_channels=out_channels, pro=pro)
        else:
            # UpCunet2x
            scale = 2
            out_channels = state_dict['unet2.conv_bottom.weight'].shape[0]
            model = UpCunet2x(in_channels=in_channels, out_channels=out_channels, pro=pro)

        return self._enhance_model(model=model, in_channels=in_channels, out_channels=out_channels, upscale=scale, name='CUGAN')



import importlib.util
import os

from ..factory import Architecture
from ..registry import Registry

internal_registry = Registry()
base_dir = os.path.dirname(__file__)
base_class = Architecture

for root, dirs, files in os.walk(base_dir):
    if root == base_dir:
        continue

    for file in files:
        if file.endswith('.py'):
            rel_dir = os.path.relpath(root, base_dir)
            module_name = os.path.join(rel_dir, file[:-3]).replace(os.path.sep, '.')
            full_module_name = f'{__name__}.{module_name}'
            module = importlib.import_module(full_module_name)

            for attribute_name in dir(module):
                attribute = getattr(module, attribute_name)
                if isinstance(attribute, type) and issubclass(attribute, base_class) and attribute is not base_class:
                    globals()[attribute_name] = attribute

                    instance = attribute()
                    internal_registry.add(instance)



from __future__ import annotations

import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from ...utilities import block as B


class RRDBNet(nn.Module):
    hyperparameters = {}

    def __init__(
        self,
        in_nc: int = 3,
        out_nc: int = 3,
        num_filters: int = 64,
        num_blocks: int = 23,
        scale: int = 4,
        plus: bool = False,
        shuffle_factor: int | None = None,
        norm=None,
        act: str = 'leakyrelu',
        upsampler: str = 'upconv',
        mode: B.ConvMode = 'CNA',
    ) -> None:
        """
        ESRGAN - Enhanced Super-Resolution Generative Adversarial Networks.
        By Xintao Wang, Ke Yu, Shixiang Wu, Jinjin Gu, Yihao Liu, Chao Dong, Yu Qiao,
        and Chen Change Loy.
        This is old-arch Residual in Residual Dense Block Network and is not
        the newest revision that's available at github.com/xinntao/ESRGAN.
        This is on purpose, the newest Network has severely limited the
        potential use of the Network with no benefits.
        This network supports model files from both new and old-arch.
        Args:
            norm: Normalization layer
            act: Activation layer
            upsampler: Upsample layer. upconv, pixel_shuffle
            mode: Convolution mode
        """
        super().__init__()
        self.shuffle_factor = shuffle_factor
        self.scale = scale // shuffle_factor if shuffle_factor else scale

        upsample_block = {
            'upconv': B.upconv_block,
            'pixel_shuffle': B.pixelshuffle_block,
        }.get(upsampler)
        if upsample_block is None:
            raise NotImplementedError(f'Upsample mode [{upsampler}] is not found')

        if scale == 3:
            upsample_blocks = upsample_block(
                in_nc=num_filters,
                out_nc=num_filters,
                upscale_factor=3,
                act_type=act,
            )
        else:
            upsample_blocks = [
                upsample_block(
                    in_nc=num_filters,
                    out_nc=num_filters,
                    act_type=act,
                )
                for _ in range(int(math.log(scale, 2)))
            ]

        self.model = B.sequential(
            # fea conv
            B.conv_block(
                in_nc=in_nc,
                out_nc=num_filters,
                kernel_size=3,
                norm_type=None,
                act_type=None,
            ),
            B.ShortcutBlock(
                B.sequential(
                    # rrdb blocks
                    *[
                        B.RRDB(
                            nf=num_filters,
                            kernel_size=3,
                            gc=32,
                            stride=1,
                            bias=True,
                            pad_type='zero',
                            norm_type=norm,
                            act_type=act,
                            mode='CNA',
                            plus=plus,
                        )
                        for _ in range(num_blocks)
                    ],
                    # lr conv
                    B.conv_block(
                        in_nc=num_filters,
                        out_nc=num_filters,
                        kernel_size=3,
                        norm_type=norm,
                        act_type=None,
                        mode=mode,
                    ),
                )
            ),
            *upsample_blocks,
            # hr_conv0
            B.conv_block(
                in_nc=num_filters,
                out_nc=num_filters,
                kernel_size=3,
                norm_type=None,
                act_type=act,
            ),
            # hr_conv1
            B.conv_block(
                in_nc=num_filters,
                out_nc=out_nc,
                kernel_size=3,
                norm_type=None,
                act_type=None,
            ),
        )

    def forward(self, x):
        if self.shuffle_factor:
            _, _, h, w = x.size()
            mod_pad_h = (self.shuffle_factor - h % self.shuffle_factor) % self.shuffle_factor
            mod_pad_w = (self.shuffle_factor - w % self.shuffle_factor) % self.shuffle_factor
            x = F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')
            x = torch.pixel_unshuffle(x, downscale_factor=self.shuffle_factor)
            x = self.model(x)
            return x[:, :, : h * self.scale, : w * self.scale]
        return self.model(x)



import torch
from torch import nn
from torch.nn import functional as F


class LayerNorm(nn.Module):
    r"""LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).
    """

    def __init__(
        self,
        normalized_shape: int,
        eps: float = 1e-6,
        data_format: str = 'channels_first',
    ):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        assert self.data_format in ['channels_last', 'channels_first']
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == 'channels_last':
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        if self.data_format == 'channels_first':
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            return self.weight[:, None, None] * x + self.bias[:, None, None]
        return None


class SADFFM(nn.Module):
    def __init__(self, dim: int, expand_ratio: float, bias: bool = True, drop: float = 0.0):
        super().__init__()
        hidden_dims = int(dim * expand_ratio)
        self.linear_in = nn.Conv2d(dim, hidden_dims * 2, kernel_size=1, bias=bias)
        self.SAL = nn.Conv2d(
            hidden_dims * 2,
            hidden_dims * 2,
            kernel_size=3,
            stride=1,
            padding=1,
            groups=hidden_dims * 2,
            bias=bias,
        )
        self.linear_out = nn.Conv2d(hidden_dims, dim, kernel_size=1, bias=bias)
        self.DFFM = DFFM(dim)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.linear_in(x)
        x1, x2 = self.SAL(x).chunk(2, dim=1)
        x = F.gelu(x1) * x2
        x = self.linear_out(x)
        x = self.DFFM(x)
        return self.drop(x)


class DFFM(nn.Module):
    def __init__(
        self,
        in_channels: int,
        act_ratio: float = 0.25,
        act_fn: type[nn.Module] = nn.GELU,
        gate_fn: type[nn.Module] = nn.Sigmoid,
    ):
        super().__init__()
        reduce_channels = int(in_channels * act_ratio)
        self.norm = LayerNorm(in_channels, data_format='channels_first')
        self.global_reduce = nn.Conv2d(in_channels, reduce_channels, 1)
        self.local_reduce = nn.Conv2d(in_channels, reduce_channels, 1)
        self.act_fn = act_fn()
        self.channel_expand = nn.Conv2d(reduce_channels, in_channels, 1)
        self.spatial_expand = nn.Conv2d(reduce_channels * 2, 1, 1)
        self.gate_fn = gate_fn()

    def forward(self, x):
        identity = x
        b, _c, _h, _w = x.shape
        x = self.norm(x)
        x_global = self.act_fn(self.global_reduce(F.adaptive_avg_pool2d(x, 1)))
        x_local = self.act_fn(self.local_reduce(x))
        c_attn = self.gate_fn(self.channel_expand(x_global))
        s_attn = self.gate_fn(self.spatial_expand(torch.cat([x_local, x_global.expand(b, -1, x.shape[2], x.shape[3])], dim=1)))
        attn = c_attn * s_attn
        return identity * attn


class Silu(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x.mul_(x.sigmoid())


class MOLRCM(nn.Module):
    def __init__(self, dim: int):
        super().__init__()

        self.split_c1, self.split_c2, self.split_c3 = (
            int((3 / 8) * dim),
            int((1 / 8) * dim),
            int((4 / 8) * dim),
        )
        self.region = nn.Conv2d(dim, dim, 5, padding=2, groups=dim)
        self.spatial_1 = nn.Conv2d(
            self.split_c1,
            self.split_c1,
            5,
            stride=1,
            padding=4,
            groups=self.split_c1,
            dilation=2,
        )
        self.spatial_2 = nn.Conv2d(
            self.split_c3,
            self.split_c3,
            7,
            stride=1,
            padding=9,
            groups=self.split_c3,
            dilation=3,
        )
        self.fusion = nn.Conv2d(dim, dim, 1)

        self.gate = Silu()
        self.proj_value = nn.Sequential(nn.Conv2d(dim, dim, 1))
        self.proj_query = nn.Sequential(nn.Conv2d(dim, dim, 1), nn.GELU())
        self.out = nn.Conv2d(dim, dim, 1)  # nn.Identity()

    def forward(self, x_):
        value = self.proj_value(x_)
        query = self.proj_query(x_)
        query = self.region(query)
        query_1 = self.spatial_1(query[:, : self.split_c1, :, :])
        query_2 = query[:, self.split_c1 : (self.split_c1 + self.split_c2), :, :]
        query_3 = self.spatial_2(query[:, (self.split_c1 + self.split_c2) :, :, :])
        out = self.gate(self.fusion(torch.cat([query_1, query_2, query_3], dim=1)))
        return self.out(out * value)


class EIMNBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float = 4.0,
        bias: bool = True,
        drop: float = 0.0,
        drop_path: float = 0.0,
        norm: type[nn.Module] = nn.BatchNorm2d,
    ):
        super().__init__()
        self.norm1 = norm(dim)
        self.attn = MOLRCM(dim)
        self.drop_path = nn.Identity()
        self.norm2 = norm(dim)
        self.mlp = SADFFM(dim, mlp_ratio, bias, drop)
        layer_scale_init_value = 1e-2
        self.layer_scale_1 = nn.Parameter(layer_scale_init_value * torch.ones(dim), requires_grad=True)
        self.layer_scale_2 = nn.Parameter(layer_scale_init_value * torch.ones(dim), requires_grad=True)

    def forward(self, x):
        x = x + self.drop_path(self.layer_scale_1.unsqueeze(-1).unsqueeze(-1) * self.attn(self.norm1(x)))
        return x + self.drop_path(self.layer_scale_2.unsqueeze(-1).unsqueeze(-1) * self.mlp(self.norm2(x)))


class eimn(nn.Module):
    def __init__(
        self,
        embed_dims: int = 64,  # detect
        scale: int = 4,
        depths: int = 1,  # detect
        mlp_ratios: float = 2.66,  # detect
        drop_rate: float = 0.0,  # detect
        drop_path_rate: float = 0.0,  # detect
        num_stages: int = 16,  # detect
        freeze_param: bool = False,
        norm: type[nn.Module] = nn.BatchNorm2d,
    ):
        super().__init__()
        depths_ = [depths] * num_stages

        self.num_stages = num_stages

        # stochastic depth decay rule
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths_))]
        cur = 0

        self.head = nn.Sequential(nn.Conv2d(3, embed_dims, 3, 1, 1))
        self.tail = nn.Sequential(nn.Conv2d(embed_dims, 3 * scale * scale, 3, 1, 1), nn.PixelShuffle(scale))

        for i in range(self.num_stages):
            block = nn.ModuleList(
                [
                    EIMNBlock(
                        dim=embed_dims,
                        mlp_ratio=mlp_ratios,
                        drop=drop_rate,
                        drop_path=dpr[cur + j],
                        norm=nn.BatchNorm2d,
                    )
                    for j in range(depths_[i])
                ]
            )
            norm = nn.LayerNorm(embed_dims)
            cur += depths_[i]

            setattr(self, f'block{i + 1}', block)
            setattr(self, f'norm{i + 1}', norm)

        if freeze_param:
            self.freeze_para()

    def freeze_param(self):
        for name, param in self.named_parameters():
            if name.split('.')[0] == 'tail':
                param.requires_grad = True
            else:
                param.requires_grad = False

    def forward(self, x):
        x = self.head(x)
        identity = x

        for i in range(self.num_stages):
            block = getattr(self, f'block{i + 1}')
            norm = getattr(self, f'norm{i + 1}')
            for blk in block:
                x = blk(x)
            x = x.permute(0, 2, 3, 1)  # (B, H, W, C)
            x = norm(x)
            x = x.permute(0, 3, 1, 2).contiguous()  # (B, C, H, W)

        return self.tail(identity + x)



import math
from typing import Literal

import torch
import torch.nn.functional as F
from torch import Tensor, nn


SampleMods = Literal['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']


class DySample(nn.Module):
    """Adapted from 'Learning to Upsample by Learning to Sample':
    https://arxiv.org/abs/2308.15085
    https://github.com/tiny-smart/dysample
    """

    def __init__(
        self,
        in_channels: int,
        out_ch: int,
        scale: int = 2,
        groups: int = 4,
        end_convolution: bool = True,
    ) -> None:
        super().__init__()

        if in_channels <= groups or in_channels % groups != 0:
            msg = 'Incorrect in_channels and groups values.'
            raise ValueError(msg)

        out_channels = 2 * groups * scale**2
        self.scale = scale
        self.groups = groups
        self.end_convolution = end_convolution
        if end_convolution:
            self.end_conv = nn.Conv2d(in_channels, out_ch, kernel_size=1)

        self.offset = nn.Conv2d(in_channels, out_channels, 1)
        self.scope = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        if self.training:
            nn.init.trunc_normal_(self.offset.weight, std=0.02)
            nn.init.constant_(self.scope.weight, val=0)

        self.register_buffer('init_pos', self._init_pos())

    def _init_pos(self) -> Tensor:
        h = torch.arange((-self.scale + 1) / 2, (self.scale - 1) / 2 + 1) / self.scale
        return torch.stack(torch.meshgrid([h, h], indexing='ij')).transpose(1, 2).repeat(1, self.groups, 1).reshape(1, -1, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        offset = self.offset(x) * self.scope(x).sigmoid() * 0.5 + self.init_pos
        B, _, H, W = offset.shape
        offset = offset.view(B, 2, -1, H, W)
        coords_h = torch.arange(H) + 0.5
        coords_w = torch.arange(W) + 0.5

        coords = (
            torch.stack(torch.meshgrid([coords_w, coords_h], indexing='ij'))
            .transpose(1, 2)
            .unsqueeze(1)
            .unsqueeze(0)
            .type(x.dtype)
            .to(x.device, non_blocking=True)
        )
        normalizer = torch.tensor([W, H], dtype=x.dtype, device=x.device, pin_memory=True).view(1, 2, 1, 1, 1)
        coords = 2 * (coords + offset) / normalizer - 1

        coords = (
            F.pixel_shuffle(coords.reshape(B, -1, H, W), self.scale)
            .view(B, 2, -1, self.scale * H, self.scale * W)
            .permute(0, 2, 3, 4, 1)
            .contiguous()
            .flatten(0, 1)
        )
        output = F.grid_sample(
            x.reshape(B * self.groups, -1, H, W),
            coords,
            mode='bilinear',
            align_corners=False,
            padding_mode='border',
        ).view(B, -1, self.scale * H, self.scale * W)

        if self.end_convolution:
            output = self.end_conv(output)

        return output


class UniUpsample(nn.Sequential):
    def __init__(
        self,
        upsample: SampleMods,
        scale: int = 2,
        in_dim: int = 64,
        out_dim: int = 3,
        mid_dim: int = 64,  # Only pixelshuffle and DySample
        group: int = 4,  # Only DySample
    ) -> None:
        m = []

        if scale == 1 or upsample == 'conv':
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'pixelshuffledirect':
            m.extend([nn.Conv2d(in_dim, out_dim * scale**2, 3, 1, 1), nn.PixelShuffle(scale)])
        elif upsample == 'pixelshuffle':
            m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            if (scale & (scale - 1)) == 0:  # scale = 2^n
                for _ in range(int(math.log2(scale))):
                    m.extend([nn.Conv2d(mid_dim, 4 * mid_dim, 3, 1, 1), nn.PixelShuffle(2)])
            elif scale == 3:
                m.extend([nn.Conv2d(mid_dim, 9 * mid_dim, 3, 1, 1), nn.PixelShuffle(3)])
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'nearest+conv':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        (
                            nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                            nn.Upsample(scale_factor=2),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        )
                    )
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            elif scale == 3:
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.Upsample(scale_factor=scale),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'dysample':
            if mid_dim != in_dim:
                m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
                dys_dim = mid_dim
            else:
                dys_dim = in_dim
            m.append(DySample(dys_dim, out_dim, scale, group))
        else:
            raise ValueError(f'An invalid Upsample was selected. Please choose one of {SampleMods}')
        super().__init__(*m)

        self.register_buffer(
            'MetaUpsample',
            torch.tensor(
                [
                    2,  # Block version, if you change something, please number from the end so that you can distinguish between authorized changes and third parties
                    list(SampleMods.__args__).index(upsample),  # UpSample method index
                    scale,
                    in_dim,
                    out_dim,
                    mid_dim,
                    group,
                ],
                dtype=torch.uint8,
            ),
        )


class FeedForward(nn.Module):
    def __init__(self, dim: int = 64, ffn_expansion_factor: float = 2.66, bias: bool = True) -> None:
        super().__init__()

        hidden_features = int(dim * ffn_expansion_factor)

        self.project_in = nn.Conv2d(dim, hidden_features * 2, kernel_size=1, bias=bias)

        self.dwconv = nn.Conv2d(
            hidden_features * 2,
            hidden_features * 2,
            kernel_size=3,
            stride=1,
            padding=1,
            groups=hidden_features * 2,
            bias=bias,
        )

        self.project_out = nn.Conv2d(hidden_features, dim, kernel_size=1, bias=bias)

    def forward(self, x):
        x = self.project_in(x)

        x1, x2 = self.dwconv(x).chunk(2, dim=1)

        x = F.gelu(x1) * x2
        x = self.project_out(x)
        return x


class LayerNorm(nn.Module):
    def __init__(self, dim: int = 64, eps: float = 1e-6) -> None:
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = eps
        self.dim = (dim,)

    def forward(self, x):
        if x.is_contiguous(memory_format=torch.channels_last):
            return F.layer_norm(x.permute(0, 2, 3, 1), self.dim, self.weight, self.bias, self.eps).permute(0, 3, 1, 2)
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight[:, None, None] * x + self.bias[:, None, None]


class DynamicLocal(nn.Module):
    def __init__(self, channels: int = 64, kernel_size: int = 3) -> None:
        super().__init__()
        self.kernel_gen = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(channels, channels, 1),
            nn.ReLU(),
            nn.Conv2d(channels, channels * kernel_size * kernel_size, 1),
        )
        self.kernel_size = kernel_size
        self.padding = kernel_size // 2

    def forward(self, x):
        B, C, H, W = x.shape
        # генерируем веса для каждого примера
        kernels = self.kernel_gen(x)  # [B, C*k*k, 1, 1]
        kernels = kernels.reshape(B * C, 1, self.kernel_size, self.kernel_size)

        # разворачиваем вход для групповой свёртки
        x_ = x.reshape(1, B * C, H, W)
        out = F.conv2d(x_, kernels, padding=self.padding, groups=B * C)
        out = out.reshape(B, C, H, W)
        return out


class FSAS(nn.Module):
    def __init__(
        self,
        dim: int = 64,
        mid_factor: float = 1.0,
        window_size: int = 8,
        windowed: bool = False,
    ) -> None:
        super().__init__()
        mid = int(dim * 3 * mid_factor)
        self.to_hidden = nn.Conv2d(dim, mid, kernel_size=1, bias=True)
        self.to_hidden_dw = nn.Conv2d(mid, mid, kernel_size=3, stride=1, padding=1, groups=mid, bias=True)

        self.project_out = nn.Conv2d(int(dim * mid_factor), dim, kernel_size=1, bias=True)
        self.patch = windowed
        self.norm = LayerNorm(int(dim * mid_factor))

        self.patch_size = window_size

    def forward(self, x):
        hidden = self.to_hidden(x)

        q, k, v = self.to_hidden_dw(hidden).chunk(3, dim=1)
        if self.patch:
            B, C, H, W = q.shape

            q_patch = q.view(
                B,
                C,
                H // self.patch_size,
                self.patch_size,
                W // self.patch_size,
                self.patch_size,
            ).permute(0, 1, 2, 4, 3, 5)  # -> [B, C, H//p, W//p, p, p]
            k_patch = k.view(
                B,
                C,
                H // self.patch_size,
                self.patch_size,
                W // self.patch_size,
                self.patch_size,
            ).permute(0, 1, 2, 4, 3, 5)

            q_fft = torch.fft.rfft2(q_patch.float())
            k_fft = torch.fft.rfft2(k_patch.float())

            out = q_fft * k_fft
            out = torch.fft.irfft2(out, s=(self.patch_size, self.patch_size))
            out = out.permute(0, 1, 2, 4, 3, 5).reshape(B, C, H, W)
        else:
            q_fft = torch.fft.rfft2(q.float())
            k_fft = torch.fft.rfft2(k.float())

            out = q_fft * k_fft
            out = torch.fft.irfft2(out)
        out = self.norm(out)

        output = v * out

        output = self.project_out(output)

        return output


class SFSAS(nn.Module):
    def __init__(
        self,
        dim: int = 64,
        split: float = 0.25,
        t_mid_factor: float = 1.0,
        window_size: int = 8,
        windowed: bool = True,
    ) -> None:
        super().__init__()
        local = int(split * dim)
        global_dim = dim - local
        self.local = nn.Sequential(DynamicLocal(local), DynamicLocal(local, 5))
        self.att = FSAS(global_dim, t_mid_factor, window_size, windowed)
        self.last = nn.Conv2d(dim, dim, 1)
        self.split = [local, global_dim]

    def forward(self, x):
        x1, x2 = x.split(self.split, dim=1)
        x1 = self.local(x1)
        x2 = self.att(x2)
        return self.last(torch.cat([x1, x2], dim=1))


class MetaBlock(nn.Module):
    def __init__(
        self,
        dim: int = 64,
        split: float = 0.25,
        t_mid_factor: float = 1.0,
        window_size: int = 8,
        windowed: bool = True,
        mlp: float = 2.66,
    ) -> None:
        super().__init__()
        self.token_mix = nn.Sequential(LayerNorm(dim), SFSAS(dim, split, t_mid_factor, window_size, windowed))
        self.channel_mix1 = nn.Sequential(LayerNorm(dim), FeedForward(dim, mlp, True))

    def forward(self, x):
        x = self.token_mix(x) + x
        x = self.channel_mix1(x) + x
        return x


class ResidualMeta(nn.Module):
    def __init__(
        self,
        dim: int = 64,
        meta_b: int = 6,
        split: float = 0.25,
        t_mid_factor: float = 1.0,
        window_size: int = 8,
        mlp: float = 2.66,
    ) -> None:
        super().__init__()
        self.residual = nn.Sequential(
            *[MetaBlock(dim, split, t_mid_factor, window_size, bool(i % 2), mlp) for i in range(meta_b)] + [DynamicLocal(dim)]
        )
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        return self.residual(x) + x


class LAWFFT(nn.Module):
    """Local Adaptive Weighted Fourier Feature Transformer"""

    def __init__(
        self,
        in_ch: int = 3,
        dim: int = 60,
        split: float = 0.25,
        scale: int = 4,
        n_rblock: int = 4,
        n_mblock: int = 6,
        t_mid_factor: float = 1.0,
        window_size: int = 8,
        mlp_factor: float = 2.66,
        unshuffle_mod: bool = False,
        upsampler: SampleMods = 'pixelshuffledirect',
        mid_dim: int = 48,
    ) -> None:
        super().__init__()
        self.scale = scale
        self.pad = 1
        if unshuffle_mod and scale < 3:
            unshuffle = 4 // scale
            self.in_to_dim = nn.Sequential(
                nn.PixelUnshuffle(unshuffle),
                nn.Conv2d(in_ch * unshuffle**2, dim, 3, 1, 1),
            )
            self.pad = unshuffle * window_size
            scale = 4
        else:
            self.in_to_dim = nn.Conv2d(in_ch, dim, 3, 1, 1)
            self.pad = window_size
        self.register_buffer('window_size', torch.tensor(window_size, dtype=torch.uint8))
        self.body = nn.Sequential(*[ResidualMeta(dim, n_mblock, split, t_mid_factor, window_size, mlp_factor) for _ in range(n_rblock)])
        self.upscale = UniUpsample(upsampler, scale, dim, in_ch, mid_dim)

        self.scale = scale

    def load_state_dict(self, state_dict, *args, **kwargs):
        state_dict['upscale.MetaUpsample'] = self.upscale.MetaUpsample
        return super().load_state_dict(state_dict, *args, **kwargs)

    def check_img_size(self, x: Tensor, resolution: tuple[int, int]) -> Tensor:
        scaled_size = self.pad
        mod_pad_h = (scaled_size - resolution[0] % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - resolution[1] % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, inp):
        b, c, h, w = inp.shape
        x = self.check_img_size(inp, (h, w))
        x = self.in_to_dim(x)
        x = self.body(x) + x
        x = self.upscale(x)[:, :, : h * self.scale, : w * self.scale]
        return x



import functools
import math
import re
from collections import OrderedDict
from typing import Mapping

from typing_extensions import override

from .arch import RRDBNet
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len


def _new_to_old_arch(state: Mapping[str, object], state_map: dict, num_blocks: int):
    """Convert a new-arch model state dictionary to an old-arch dictionary."""
    if 'params_ema' in state:
        state = state['params_ema']

    if 'conv_first.weight' not in state:
        # model is already old arch, this is a loose check, but should be sufficient
        return state

    # add nb to state keys
    for kind in ('weight', 'bias'):
        state_map[f'model.1.sub.{num_blocks}.{kind}'] = state_map[f'model.1.sub./NB/.{kind}']
        del state_map[f'model.1.sub./NB/.{kind}']

    old_state = OrderedDict()
    for old_key, new_keys in state_map.items():
        for new_key in new_keys:
            if r'\1' in old_key:
                for k, v in state.items():
                    sub = re.sub(new_key, old_key, k)
                    if sub != k:
                        old_state[sub] = v
            else:
                if new_key in state:
                    old_state[old_key] = state[new_key]

    # upconv layers
    max_upconv = 0
    for key in state.keys():
        match = re.match(r'(upconv|conv_up)(\d)\.(weight|bias)', key)
        if match is not None:
            _, key_num, key_type = match.groups()
            old_state[f'model.{int(key_num) * 3}.{key_type}'] = state[key]
            max_upconv = max(max_upconv, int(key_num) * 3)

    # final layers
    for key in state.keys():
        if key in ('HRconv.weight', 'conv_hr.weight'):
            old_state[f'model.{max_upconv + 2}.weight'] = state[key]
        elif key in ('HRconv.bias', 'conv_hr.bias'):
            old_state[f'model.{max_upconv + 2}.bias'] = state[key]
        elif key in ('conv_last.weight',):
            old_state[f'model.{max_upconv + 4}.weight'] = state[key]
        elif key in ('conv_last.bias',):
            old_state[f'model.{max_upconv + 4}.bias'] = state[key]

    # Sort by first numeric value of each layer
    def compare(item1: str, item2: str):
        parts1 = item1.split('.')
        parts2 = item2.split('.')
        int1 = int(parts1[1])
        int2 = int(parts2[1])
        return int1 - int2

    sorted_keys = sorted(old_state.keys(), key=functools.cmp_to_key(compare))

    # Rebuild the output dict in the right order
    out_dict = OrderedDict((k, old_state[k]) for k in sorted_keys)

    return out_dict


def _get_scale(state: Mapping[str, object]) -> int:
    # model is composed of a few blocks that look like this flattened:
    #
    #   Conv2d
    #   B.ShortcutBlock
    #   [nn.Upsample, Conv2d, nn.LeakyReLU] for i in range(log2(scale))
    #   Conv2d
    #   nn.LeakyReLU (activation)
    #   Conv2d
    seq_len = get_seq_len(state, 'model')
    log2_scale = (seq_len - 5) // 3
    return 2**log2_scale


def _get_num_blocks(state: Mapping[str, object], state_map: dict) -> int:
    nbs = []
    state_keys = state_map[r'model.1.sub.\1.RDB\2.conv\3.0.\4'] + (r'model\.\d+\.sub\.(\d+)\.RDB(\d+)\.conv(\d+)\.0\.(weight|bias)',)
    for state_key in state_keys:
        for k in state:
            m = re.search(state_key, k)
            if m:
                nbs.append(int(m.group(1)))
        if nbs:
            break
    return max(*nbs) + 1


def _to_old_arch(state: Mapping[str, object]) -> Mapping[str, object]:
    state_map = {
        # currently supports old, new, and newer RRDBNet arch models
        # ESRGAN, BSRGAN/RealSR, Real-ESRGAN
        'model.0.weight': ('conv_first.weight',),
        'model.0.bias': ('conv_first.bias',),
        'model.1.sub./NB/.weight': ('trunk_conv.weight', 'conv_body.weight'),
        'model.1.sub./NB/.bias': ('trunk_conv.bias', 'conv_body.bias'),
        r'model.1.sub.\1.RDB\2.conv\3.0.\4': (
            r'RRDB_trunk\.(\d+)\.RDB(\d)\.conv(\d+)\.(weight|bias)',
            r'body\.(\d+)\.rdb(\d)\.conv(\d+)\.(weight|bias)',
        ),
    }

    if 'params_ema' in state:
        state = state['params_ema']

    num_blocks = _get_num_blocks(state, state_map)
    return _new_to_old_arch(state, state_map, num_blocks)


class ESRGANArch(Architecture[RRDBNet]):
    def __init__(self) -> None:
        super().__init__(
            uid='ESRGAN',
            detect=KeyCondition.has_any(
                KeyCondition.has_all(
                    'model.0.weight',
                    'model.1.sub.0.RDB1.conv1.0.weight',
                ),
                KeyCondition.has_all(
                    'conv_first.weight',
                    'body.0.rdb1.conv1.weight',
                    'conv_body.weight',
                    'conv_last.weight',
                ),
                KeyCondition.has_all(
                    # BSRGAN/RealSR
                    'conv_first.weight',
                    'RRDB_trunk.0.RDB1.conv1.weight',
                    'trunk_conv.weight',
                    'conv_last.weight',
                ),
                KeyCondition.has_all(
                    # ESRGAN+
                    'model.0.weight',
                    'model.1.sub.0.RDB1.conv1x1.weight',
                ),
            ),
        )

    @override
    def load(self, state_dict: Mapping[str, object]):
        # default values
        plus: bool = False

        state_dict = _to_old_arch(state_dict)

        model_seq_len = get_seq_len(state_dict, 'model')

        in_nc = state_dict['model.0.weight'].shape[1]
        out_nc = state_dict[f'model.{model_seq_len - 1}.weight'].shape[0]

        scale = _get_scale(state_dict)
        num_blocks = get_seq_len(state_dict, 'model.1.sub') - 1
        num_filters = state_dict['model.0.weight'].shape[0]

        if any('.conv1x1.' in k for k in state_dict.keys()):
            plus = True

        # Detect if pixelunshuffle was used (Real-ESRGAN)
        if in_nc in (out_nc * 4, out_nc * 16) and out_nc in (
            in_nc / 4,
            in_nc / 16,
        ):
            shuffle_factor = int(math.sqrt(in_nc / out_nc))
        else:
            shuffle_factor = None

        model = RRDBNet(
            in_nc=in_nc,
            out_nc=out_nc,
            num_filters=num_filters,
            num_blocks=num_blocks,
            scale=scale,
            plus=plus,
            shuffle_factor=shuffle_factor,
        )
        if shuffle_factor:
            in_nc //= shuffle_factor**2
            scale //= shuffle_factor
        return self._enhance_model(model=model, in_channels=in_nc, out_channels=out_nc, upscale=scale, name='ESRGAN')



from typing import Mapping

from .arch import SpanPlus
from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import get_seq_len, pixelshuffle_scale, dysample_scale


class SpanPlusArch(Architecture[SpanPlus]):
    def __init__(self):
        super().__init__(
            uid='spanplus',
            detect=KeyCondition.has_all('feats.0.eval_conv.weight'),
        )

    def load(self, state_dict: Mapping[str, object]):
        n_feats = get_seq_len(state_dict, 'feats') - 1
        blocks = [get_seq_len(state_dict, f'feats.{n_feat + 1}.block_n') for n_feat in range(n_feats)]
        num_in_ch = state_dict['feats.0.eval_conv.weight'].shape[1]
        feature_channels = state_dict['feats.0.eval_conv.weight'].shape[0]
        if 'upsampler.0.weight' in state_dict.keys():
            upsampler = 'ps'
            num_out_ch = num_in_ch
            upscale = pixelshuffle_scale(state_dict['upsampler.0.weight'].shape[0], num_out_ch)
        else:
            upsampler = 'dys'
            num_out_ch = state_dict['upsampler.end_conv.weight'].shape[0]
            upscale = dysample_scale(state_dict['upsampler.offset.weight'].shape[0])

        model = SpanPlus(
            num_in_ch=num_in_ch,
            num_out_ch=num_out_ch,
            blocks=blocks,
            feature_channels=feature_channels,
            upscale=upscale,
            upsampler=upsampler,
        )

        return self._enhance_model(model=model, in_channels=num_in_ch, out_channels=num_out_ch, upscale=upscale, name='SPANPlus')
